# Tests for training
# hardware config: cpu / single gpu / multi gpu
# tests (single gpu):
#   - basic training
#   - optimizer
#   - net-g
#   - net-gs
#   - with-val
# tests (multi gpu):
#   - basic training

import subprocess
import os
import torch
import pytest

available_gpus = torch.cuda.device_count()
DDP_PARAM = '--use-torchrun "-t 3 --log_dir ../log/ --nproc_per_node 1"'
CMD_BASIC = 'python cli.py trainlaunch --model {model} --dataroot {dataroot} --name test_local --modalities-no {modalities_no} --seg-gen {seg_gen} --batch-size 1 --num-threads 0 --checkpoints-dir {dir_save} --remote True --n-epochs 1 --n-epochs-decay 1'
CMD_KD = ' --model-dir-teacher {model_dir_teacher}'

print('Building and installing deepliif locally to test trainlaunch..')
res = subprocess.run('pip install -e . --no-dependencies',shell=True)
assert res.returncode == 0

#----------------------
#---- gpu: single -----
#----------------------
def test_cli_trainlaunch_single_gpu(tmp_path, model_info, foldername_suffix):
    if available_gpus > 0:
        torch.cuda.nvtx.range_push("test_cli_trainlaunch_single_gpu")
        dirs_input = model_info['dir_input_train']
        
        for i in range(len(dirs_input)):
            torch.cuda.nvtx.range_push(f"test_cli_trainlaunch_single_gpu {dirs_input[i]}")
            dir_save = tmp_path
            
            fns_input = [f for f in os.listdir(dirs_input[i] + '/train' + foldername_suffix) if os.path.isfile(os.path.join(dirs_input[i] + '/train' + foldername_suffix, f)) and f.endswith('png')]
            num_input = len(fns_input)
            assert num_input > 0
    
            test_param = '--gpu-ids 0'
            cmd = CMD_BASIC.format(model=model_info["model"], dataroot=dirs_input[i], 
                                   modalities_no=model_info["modalities_no"][i], 
                                   seg_gen=model_info["seg_gen"][i], dir_save=dir_save)
            cmd += f' {test_param}'
            if model_info["model"] in ['DeepLIIFKD']:
                cmd += CMD_KD.format(model_dir_teacher=model_info['model_dir_teacher'][i])
            cmd += f' {DDP_PARAM}'
            res = subprocess.run(cmd,shell=True)
            assert res.returncode == 0
            torch.cuda.nvtx.range_pop()
        torch.cuda.nvtx.range_pop()
    else:
        pytest.skip(f'Detected {available_gpus} (< 1) available GPUs. Skip.')


def test_cli_trainlaunch_single_gpu_optimizer(tmp_path, model_info, foldername_suffix):
    if available_gpus > 0:
        torch.cuda.nvtx.range_push("test_cli_trainlaunch_single_gpu_optimizer")
        dirs_input = model_info['dir_input_train']
        for i in range(len(dirs_input)):
            torch.cuda.nvtx.range_push(f"test_cli_trainlaunch_single_gpu {dirs_input[i]}")
            dir_save = tmp_path
            
            
            fns_input = [f for f in os.listdir(dirs_input[i] + '/train' + foldername_suffix) if os.path.isfile(os.path.join(dirs_input[i] + '/train' + foldername_suffix, f)) and f.endswith('png')]
            num_input = len(fns_input)
            assert num_input > 0
    
            test_param = '--gpu-ids 0 --optimizer sgd'
            cmd = CMD_BASIC.format(model=model_info["model"], dataroot=dirs_input[i], 
                                   modalities_no=model_info["modalities_no"][i], 
                                   seg_gen=model_info["seg_gen"][i], dir_save=dir_save)
            cmd += f' {test_param}'
            if model_info["model"] in ['DeepLIIFKD']:
                cmd += CMD_KD.format(model_dir_teacher=model_info['model_dir_teacher'][i])
            cmd += f' {DDP_PARAM}'
            res = subprocess.run(cmd,shell=True)
            assert res.returncode == 0
            torch.cuda.nvtx.range_pop()
        torch.cuda.nvtx.range_pop()
    else:
        pytest.skip(f'Detected {available_gpus} (< 1) available GPUs. Skip.')


def test_cli_trainlaunch_single_gpu_netg(tmp_path, model_info, foldername_suffix):
    if available_gpus > 0:
        torch.cuda.nvtx.range_push("test_cli_trainlaunch_single_gpu_netg")
        dirs_input = model_info['dir_input_train']
        for i in range(len(dirs_input)):
            torch.cuda.nvtx.range_push(f"test_cli_trainlaunch_single_gpu {dirs_input[i]}")
            dir_save = tmp_path
            
            fns_input = [f for f in os.listdir(dirs_input[i] + '/train' + foldername_suffix) if os.path.isfile(os.path.join(dirs_input[i] + '/train' + foldername_suffix, f)) and f.endswith('png')]
            num_input = len(fns_input)
            assert num_input > 0
    
            test_param = '--gpu-ids 0 --net-g unet_512'
            cmd = CMD_BASIC.format(model=model_info["model"], dataroot=dirs_input[i], 
                                   modalities_no=model_info["modalities_no"][i], 
                                   seg_gen=model_info["seg_gen"][i], dir_save=dir_save)
            cmd += f' {test_param}'
            if model_info["model"] in ['DeepLIIFKD']:
                cmd += CMD_KD.format(model_dir_teacher=model_info['model_dir_teacher'][i])
            cmd += f' {DDP_PARAM}'
            res = subprocess.run(cmd,shell=True)
            assert res.returncode == 0
            
            test_param = '--gpu-ids 0 --net-g unet_512_attention'
            cmd = CMD_BASIC.format(model=model_info["model"], dataroot=dirs_input[i], 
                                   modalities_no=model_info["modalities_no"][i], 
                                   seg_gen=model_info["seg_gen"][i], dir_save=dir_save)
            cmd += f' {test_param}'
            if model_info["model"] in ['DeepLIIFKD']:
                cmd += CMD_KD.format(model_dir_teacher=model_info['model_dir_teacher'][i])
            cmd += f' {DDP_PARAM}'
            res = subprocess.run(cmd,shell=True)
            assert res.returncode == 0
            torch.cuda.nvtx.range_pop()
        torch.cuda.nvtx.range_pop()
    else:
        pytest.skip(f'Detected {available_gpus} (< 1) available GPUs. Skip.')
        

def test_cli_trainlaunch_single_gpu_netgs(tmp_path, model_info, foldername_suffix):
    if available_gpus > 0:
        torch.cuda.nvtx.range_push("test_cli_trainlaunch_single_gpu_netgs")
        dirs_input = model_info['dir_input_train']
        for i in range(len(dirs_input)):
            torch.cuda.nvtx.range_push(f"test_cli_trainlaunch_single_gpu {dirs_input[i]}")
            dir_save = tmp_path
            
            fns_input = [f for f in os.listdir(dirs_input[i] + '/train' + foldername_suffix) if os.path.isfile(os.path.join(dirs_input[i] + '/train' + foldername_suffix, f)) and f.endswith('png')]
            num_input = len(fns_input)
            assert num_input > 0
    
            test_param = '--gpu-ids 0 --net-gs unet_512'
            cmd = CMD_BASIC.format(model=model_info["model"], dataroot=dirs_input[i], 
                                   modalities_no=model_info["modalities_no"][i], 
                                   seg_gen=model_info["seg_gen"][i], dir_save=dir_save)
            cmd += f' {test_param}'
            if model_info["model"] in ['DeepLIIFKD']:
                cmd += CMD_KD.format(model_dir_teacher=model_info['model_dir_teacher'][i])
            cmd += f' {DDP_PARAM}'
            res = subprocess.run(cmd,shell=True)
            assert res.returncode == 0
            
            test_param = '--gpu-ids 0 --net-gs unet_512_attention'
            cmd = CMD_BASIC.format(model=model_info["model"], dataroot=dirs_input[i], 
                                   modalities_no=model_info["modalities_no"][i], 
                                   seg_gen=model_info["seg_gen"][i], dir_save=dir_save)
            cmd += f' {test_param}'
            if model_info["model"] in ['DeepLIIFKD']:
                cmd += CMD_KD.format(model_dir_teacher=model_info['model_dir_teacher'][i])
            cmd += f' {DDP_PARAM}'
            res = subprocess.run(cmd,shell=True)
            assert res.returncode == 0
            torch.cuda.nvtx.range_pop()
        torch.cuda.nvtx.range_pop()
    else:
        pytest.skip(f'Detected {available_gpus} (< 1) available GPUs. Skip.')


def test_cli_trainlaunch_single_gpu_withval(tmp_path, model_info, foldername_suffix):
    if available_gpus > 0 and model_info["model"] not in ['CycleGAN']:
        torch.cuda.nvtx.range_push("test_cli_trainlaunch_single_gpu_withval")
        dirs_input = model_info['dir_input_train']
        for i in range(len(dirs_input)):
            torch.cuda.nvtx.range_push(f"test_cli_trainlaunch_single_gpu {dirs_input[i]}")
            dir_save = tmp_path
            
            fns_input = [f for f in os.listdir(dirs_input[i] + '/train' + foldername_suffix) if os.path.isfile(os.path.join(dirs_input[i] + '/train' + foldername_suffix, f)) and f.endswith('png')]
            num_input = len(fns_input)
            assert num_input > 0
    
            test_param = '--gpu-ids 0 --with-val'
            cmd = CMD_BASIC.format(model=model_info["model"], dataroot=dirs_input[i], 
                                   modalities_no=model_info["modalities_no"][i], 
                                   seg_gen=model_info["seg_gen"][i], dir_save=dir_save)
            cmd += f' {test_param}'
            if model_info["model"] in ['DeepLIIFKD']:
                cmd += CMD_KD.format(model_dir_teacher=model_info['model_dir_teacher'][i])
            cmd += f' {DDP_PARAM}'
            res = subprocess.run(cmd,shell=True)
            assert res.returncode == 0
            torch.cuda.nvtx.range_pop()
        torch.cuda.nvtx.range_pop()
    else:
        pytest.skip(f'Detected {available_gpus} (< 1) available GPUs. Skip.')

#----------------------
#---- gpu: multi ------
#----------------------
def test_cli_trainlaunch_multi_gpu_dp(tmp_path, model_info, foldername_suffix):
    if available_gpus > 1:
        torch.cuda.nvtx.range_push("test_cli_trainlaunch_multi_gpu_dp")
        dirs_input = model_info['dir_input_train']
        for i in range(len(dirs_input)):
            torch.cuda.nvtx.range_push(f"test_cli_trainlaunch_multi_gpu_dp {dirs_input[i]}")
            dir_save = tmp_path
            
            fns_input = [f for f in os.listdir(dirs_input[i] + '/train' + foldername_suffix) if os.path.isfile(os.path.join(dirs_input[i] + '/train' + foldername_suffix, f)) and f.endswith('png')]
            num_input = len(fns_input)
            assert num_input > 0
            
            test_param = '--gpu-ids 0 --gpu-ids 1'
            cmd = CMD_BASIC.format(model=model_info["model"], dataroot=dirs_input[i], 
                                   modalities_no=model_info["modalities_no"][i], 
                                   seg_gen=model_info["seg_gen"][i], dir_save=dir_save)
            cmd += f' {test_param}'
            if model_info["model"] in ['DeepLIIFKD']:
                cmd += CMD_KD.format(model_dir_teacher=model_info['model_dir_teacher'][i])
            cmd += f' {DDP_PARAM}'
            res = subprocess.run(cmd,shell=True)
            assert res.returncode == 0
            torch.cuda.nvtx.range_pop()
        torch.cuda.nvtx.range_pop()
    else:
        pytest.skip(f'Detected {available_gpus} (< 2) available GPUs. Skip.')




# adapted from https://github.com/LeeJunHyun/Image_Segmentation/blob/master/network.py

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn import init

def init_weights(net, init_type='normal', gain=0.02):
    def init_func(m):
        classname = m.__class__.__name__
        if hasattr(m, 'weight') and (classname.find('Conv') != -1 or classname.find('Linear') != -1):
            if init_type == 'normal':
                init.normal_(m.weight.data, 0.0, gain)
            elif init_type == 'xavier':
                init.xavier_normal_(m.weight.data, gain=gain)
            elif init_type == 'kaiming':
                init.kaiming_normal_(m.weight.data, a=0, mode='fan_in')
            elif init_type == 'orthogonal':
                init.orthogonal_(m.weight.data, gain=gain)
            else:
                raise NotImplementedError('initialization method [%s] is not implemented' % init_type)
            if hasattr(m, 'bias') and m.bias is not None:
                init.constant_(m.bias.data, 0.0)
        elif classname.find('BatchNorm2d') != -1:
            init.normal_(m.weight.data, 1.0, gain)
            init.constant_(m.bias.data, 0.0)

    print('initialize network with %s' % init_type)
    net.apply(init_func)

class conv_block(nn.Module):
    def __init__(self,ch_in,ch_out,innermost=False,outermost=False):
        super(conv_block,self).__init__()
        if outermost:
            self.conv = nn.Sequential(
                nn.Conv2d(ch_in, ch_out, kernel_size=4,stride=2,padding=1,bias=True),
                nn.LeakyReLU(0.2, True),
            )
        elif innermost:
            self.conv = nn.Sequential(
                nn.Conv2d(ch_in, ch_out, kernel_size=4,stride=2,padding=1,bias=True),
                nn.ReLU(inplace=True),
            )
        else:
            self.conv = nn.Sequential(
                nn.Conv2d(ch_in, ch_out, kernel_size=4,stride=2,padding=1,bias=True),
                nn.BatchNorm2d(ch_out),
                nn.LeakyReLU(0.2, True),
            )

    def forward(self,x):
        x = self.conv(x)
        return x

class up_conv(nn.Module):
    def __init__(self,ch_in,ch_out,innermost=False,outermost=False):
        super(up_conv,self).__init__()
        use_bias=False
        if outermost:
            self.up = nn.Sequential(
                          nn.ConvTranspose2d(ch_in * 2, ch_out,
                                        kernel_size=4, stride=2,
                                        padding=1),
                          nn.Tanh())
        elif innermost:
            self.up = nn.Sequential(
                          nn.ConvTranspose2d(ch_in, ch_out,
                                        kernel_size=4, stride=2,
                                        padding=1, bias=use_bias),
                          nn.BatchNorm2d(ch_out),
                          nn.ReLU(True))
        else:
            self.up = nn.Sequential(
                          nn.ConvTranspose2d(ch_in * 2, ch_out,
                                        kernel_size=4, stride=2,
                                        padding=1, bias=use_bias),
                          nn.BatchNorm2d(ch_out),
                          nn.ReLU(True))
                                        

    def forward(self,x):
        x = self.up(x)
        return x


class Attention_block(nn.Module):
    def __init__(self,F_g,F_l,F_int):
        super(Attention_block,self).__init__()
        self.W_g = nn.Sequential(
            nn.Conv2d(F_g, F_int, kernel_size=1,stride=1,padding=0,bias=True),
            nn.BatchNorm2d(F_int)
            )
        
        self.W_x = nn.Sequential(
            nn.Conv2d(F_l, F_int, kernel_size=1,stride=1,padding=0,bias=True),
            nn.BatchNorm2d(F_int)
        )

        self.psi = nn.Sequential(
            nn.Conv2d(F_int, 1, kernel_size=1,stride=1,padding=0,bias=True),
            nn.BatchNorm2d(1),
            nn.Sigmoid()
        )
        
        self.relu = nn.ReLU(inplace=True)
        
    def forward(self,g,x):
        g1 = self.W_g(g)
        x1 = self.W_x(x)
        psi = self.relu(g1+x1)
        psi = self.psi(psi)

        return x*psi



class AttU_Net(nn.Module):
    def __init__(self,img_ch=3,output_ch=1):
        super(AttU_Net,self).__init__()
        
        self.Conv1 = conv_block(ch_in=img_ch,ch_out=64,outermost=True)
        self.Conv2 = conv_block(ch_in=64,ch_out=128)
        self.Conv3 = conv_block(ch_in=128,ch_out=256)
        self.Conv4 = conv_block(ch_in=256,ch_out=512)
        self.Conv5 = conv_block(ch_in=512,ch_out=512)
        self.Conv6 = conv_block(ch_in=512,ch_out=512)
        self.Conv7 = conv_block(ch_in=512,ch_out=512)
        self.Conv8 = conv_block(ch_in=512,ch_out=512,innermost=True)
        #self.Conv9 = conv_block(ch_in=512,ch_out=512,innermost=True)
        
        self.Up8 = up_conv(ch_in=512,ch_out=512,innermost=True)
        self.Att8 = Attention_block(F_g=512,F_l=512,F_int=512)
        
        self.Up7 = up_conv(ch_in=512,ch_out=512)
        self.Att7 = Attention_block(F_g=512,F_l=512,F_int=512)
        
        self.Up6 = up_conv(ch_in=512,ch_out=512)
        self.Att6 = Attention_block(F_g=512,F_l=512,F_int=512)
        
        self.Up5 = up_conv(ch_in=512,ch_out=512)
        self.Att5 = Attention_block(F_g=512,F_l=512,F_int=512)
        
        self.Up4 = up_conv(ch_in=512,ch_out=256)
        self.Att4 = Attention_block(F_g=256,F_l=256,F_int=128)
        
        self.Up3 = up_conv(ch_in=256,ch_out=128)
        self.Att3 = Attention_block(F_g=128,F_l=128,F_int=64)
        
        self.Up2 = up_conv(ch_in=128,ch_out=64)
        self.Att2 = Attention_block(F_g=64,F_l=64,F_int=32)
        
        self.Up1 = up_conv(ch_in=64,ch_out=output_ch,outermost=True)

    def forward(self,x):
        # encoding path
        x1 = self.Conv1(x)
        x2 = self.Conv2(x1)
        x3 = self.Conv3(x2)
        x4 = self.Conv4(x3)
        x5 = self.Conv5(x4)
        x6 = self.Conv6(x5)
        x7 = self.Conv7(x6)
        x8 = self.Conv8(x7)
        #x9 = self.Conv9(x8)
        
        #d9 = self.Up
        d8 = self.Up8(x8)
        x7 = self.Att8(g=d8,x=x7)
        d8 = torch.cat((x7,d8),dim=1)

        d7 = self.Up7(d8)
        x6 = self.Att7(g=d7,x=x6)
        d7 = torch.cat((x6,d7),dim=1)
        
        d6 = self.Up6(d7)
        x5 = self.Att6(g=d6,x=x5)
        d6 = torch.cat((x5,d6),dim=1)
        
        d5 = self.Up5(d6)
        x4 = self.Att5(g=d5,x=x4)
        d5 = torch.cat((x4,d5),dim=1)
        
        
        d4 = self.Up4(d5) # x4: [2, 512, 4, 4], d4: [2, 256, 4, 4]
        x3 = self.Att4(g=d4,x=x3)
        d4 = torch.cat((x3,d4),dim=1)

        d3 = self.Up3(d4)
        x2 = self.Att3(g=d3,x=x2)
        d3 = torch.cat((x2,d3),dim=1)

        d2 = self.Up2(d3)
        x1 = self.Att2(g=d2,x=x1)
        d2 = torch.cat((x1,d2),dim=1)

        d1 = self.Up1(d2)

        return d1




"""This module contains simple helper functions """
import os
from time import time
from functools import wraps

import torch
import numpy as np
from PIL import Image
import cv2
from skimage.metrics import structural_similarity as ssim


def timeit(f):
    @wraps(f)
    def wrap(*args, **kwargs):
        ts = time()
        result = f(*args, **kwargs)
        print(f'{f.__name__} {time() - ts}')

        return result

    return wrap


def diagnose_network(net, name='network'):
    """Calculate and print the mean of average absolute(gradients)

    Parameters:
        net (torch network) -- Torch network
        name (str) -- the name of the network
    """
    mean = 0.0
    count = 0
    for param in net.parameters():
        if param.grad is not None:
            mean += torch.mean(torch.abs(param.grad.data))
            count += 1
    if count > 0:
        mean = mean / count
    print(name)
    print(mean)


def save_image(image_numpy, image_path, aspect_ratio=1.0):
    """Save a numpy image to the disk

    Parameters:
        image_numpy (numpy array) -- input numpy array
        image_path (str)          -- the path of the image
    """
    x, y, nc = image_numpy.shape
    
    if nc > 3:
        if nc % 3 == 0:
            nc_img = 3
            no_img = nc // nc_img
            
        elif nc % 2 == 0:
            nc_img = 2
            no_img = nc // nc_img
        else:
            nc_img = 1
            no_img = nc // nc_img
        print(f'image (numpy) has {nc}>3 channels, inferred to have {no_img} images each with {nc_img} channel(s)')
        l_image_numpy = np.dsplit(image_numpy,[nc_img*i for i in range(1,no_img)])
        image_numpy = np.concatenate(l_image_numpy, axis=1) # stack horizontally
        
    image_pil = Image.fromarray(image_numpy)
    h, w, _ = image_numpy.shape

    if aspect_ratio > 1.0:
        image_pil = image_pil.resize((h, int(w * aspect_ratio)), Image.BICUBIC)
    if aspect_ratio < 1.0:
        image_pil = image_pil.resize((int(h / aspect_ratio), w), Image.BICUBIC)
    image_pil.save(image_path)


def print_numpy(x, val=True, shp=False):
    """Print the mean, min, max, median, std, and size of a numpy array

    Parameters:
        val (bool) -- if print the values of the numpy array
        shp (bool) -- if print the shape of the numpy array
    """
    x = x.astype(np.float64)
    if shp:
        print('shape,', x.shape)
    if val:
        x = x.flatten()
        print('mean = %3.3f, min = %3.3f, max = %3.3f, median = %3.3f, std=%3.3f' % (
            np.mean(x), np.min(x), np.max(x), np.median(x), np.std(x)))


def mkdirs(paths):
    """create empty directories if they don't exist

    Parameters:
        paths (str list) -- a list of directory paths
    """
    if isinstance(paths, list) and not isinstance(paths, str):
        for path in paths:
            mkdir(path)
    else:
        mkdir(paths)


def mkdir(path):
    """create a single empty directory if it didn't exist

    Parameters:
        path (str) -- a single directory path
    """
    if not os.path.exists(path):
        os.makedirs(path, exist_ok=True)


def tensor2im(input_image, imtype=np.uint8):
    """"Converts a Tensor array into a numpy image array.

    Parameters:
        input_image (tensor) --  the input image tensor array
        imtype (type)        --  the desired type of the converted numpy array
    """
    if not isinstance(input_image, np.ndarray):
        if isinstance(input_image, torch.Tensor):  # get the data from a variable
            image_tensor = input_image.data
        else:
            return input_image
        image_numpy = image_tensor[0].cpu().float().numpy()  # convert it into a numpy array
        if image_numpy.shape[0] == 1:  # grayscale to RGB
            image_numpy = np.tile(image_numpy, (3, 1, 1))
        image_numpy = (np.transpose(image_numpy, (1, 2, 0)) + 1) / 2.0 * 255.0  # post-processing: tranpose and scaling
    else:  # if it is a numpy array, do nothing
        image_numpy = input_image
    return image_numpy.astype(imtype)


def tensor_to_pil(t):
    return Image.fromarray(tensor2im(t))


def calculate_ssim(img1, img2):
    return ssim(img1, img2, data_range=img2.max() - img2.min())


def check_multi_scale(img1, img2):
    img1 = np.array(img1)
    img2 = np.array(img2)
    max_ssim = (512, 0)
    for tile_size in range(100, 1000, 100):
        image_ssim = 0
        tile_no = 0
        for i in range(0, img2.shape[0], tile_size):
            for j in range(0, img2.shape[1], tile_size):
                if i + tile_size <= img2.shape[0] and j + tile_size <= img2.shape[1]:
                    tile = img2[i: i + tile_size, j: j + tile_size]
                    tile = cv2.resize(tile, (img1.shape[0], img1.shape[1]))
                    tile_ssim = calculate_ssim(img1, tile)
                    image_ssim += tile_ssim
                    tile_no += 1
        if tile_no > 0:
            image_ssim /= tile_no
            if max_ssim[1] < image_ssim:
                max_ssim = (tile_size, image_ssim)
    return max_ssim[0]


import subprocess
import os
from threading import Thread , Timer
import sched, time

# modified from https://stackoverflow.com/questions/67707828/how-to-get-every-seconds-gpu-usage-in-python
def get_gpu_memory(gpu_id=0):
    """
    Currently collects gpu memory info for a given gpu id.
    """
    output_to_list = lambda x: x.decode('ascii').split('\n')[:-1]
    ACCEPTABLE_AVAILABLE_MEMORY = 1024
    COMMAND = "nvidia-smi --query-gpu=memory.used --format=csv"
    try:
        memory_use_info = output_to_list(subprocess.check_output(COMMAND.split(),stderr=subprocess.STDOUT))[1:]
    except subprocess.CalledProcessError as e:
        raise RuntimeError("command '{}' return with error (code {}): {}".format(e.cmd, e.returncode, e.output))
    memory_use_values = [int(x.split()[0]) for i, x in enumerate(memory_use_info)]
    
    #assert len(memory_use_values)==1, f"get_gpu_memory::memory_use_values should have only 1 value, now has {len(memory_use_values)} (memory_use_values)"
    return memory_use_values[gpu_id]

class HardwareStatus():
    def __init__(self):
        self.gpu_mem = []
        self.timer = None
    
    def get_status_every_sec(self, gpu_id=0):
        """
            This function calls itself every 1 sec and appends the gpu_memory.
        """
        self.timer = Timer(1.0, self.get_status_every_sec)
        self.timer.start()
        self.gpu_mem.append(get_gpu_memory(gpu_id))
        # print('self.gpu_mem',self.gpu_mem)
    
    def stop_timer(self):
        self.timer.cancel()


def get_mod_id_seg(dir_model):
    # assume we already know there are seg models - this check is intended to be done prior to calling this function
    fns = [fn for fn in os.listdir(dir_model) if fn.endswith('.pth') and 'net_G' in fn]
    
    if len(fns) == 0: # typically this means the directory only contains serialized models
        fns = [fn for fn in os.listdir(dir_model) if fn.endswith('.pt') and fn.startswith('G')]
        model_names = [fn[1:-3] for fn in fns] # 1[:-3] drop ".pt" and the starting G
    else:
        model_names = [fn[:-4].split('_')[2][1:] for fn in fns] # [1:] drop "G"
    
    if len(fns) == 0:
        raise Exception('Cannot find any model file ending with .pt or .pth in directory',dir_model)
    
    model_name_seg = max(model_names, key=len)
    return model_name_seg[0]

def get_input_id(dir_model):
    # assume we already know there are seg models - this check is intended to be done prior to calling this function
    fns = [fn for fn in os.listdir(dir_model) if fn.endswith('.pth') and 'net_G' in fn]
    
    if len(fns) == 0: # typically this means the directory only contains serialized models
        fns = [fn for fn in os.listdir(dir_model) if fn.endswith('.pt') and fn.startswith('G')]
        model_names_seg = [fn[2:-3] for fn in fns] # [2:] drop "GS"/"G5"
    else:
        model_names_seg = [fn[:-4].split('_')[2][2:] for fn in fns] # [2:] drop "GS"/"G5"
    
    if len(fns) == 0:
        raise Exception('Cannot find any model file ending with .pt or .pth in directory',dir_model)
      
    if '0' in model_names_seg:
        return '0'
    else:
        return '1'
    
def init_input_and_mod_id(opt, dir_model=None):
    """
    Used by model classes to initialize input id and mod id under different situations.
    """
    mod_id_seg = None
    input_id = None
    
    if not opt.continue_train and opt.is_train:
        if hasattr(opt,'mod_id_seg'): # use mod id seg from train opt file if available
            mod_id_seg = opt.mod_id_seg
        elif not hasattr(opt,'modalities_names'): # backward compatible with models trained before this param was introduced
            mod_id_seg = opt.modalities_no + 1 # for original DeepLIIF, modalities_no is 4 and the seg mod id is 5
        else:
            mod_id_seg = 'S'
        
        if opt.model in ['DeepLIIF','DeepLIIFKD']:
            input_id = '0'
    else: # for both continue train and test, we load existing models, so need to obtain seg mod id and input id from filenames
        if hasattr(opt, 'mod_id_seg'):
            mod_id_seg = opt.mod_id_seg
        else:
            # for contiue-training, extract mod id seg from existing files if not available
            mod_id_seg = get_mod_id_seg(dir_model if dir_model is not None else os.path.join(opt.checkpoints_dir, opt.name))
        
        if opt.model in ['DeepLIIF','DeepLIIFKD']:
            input_id = get_input_id(dir_model if dir_model is not None else os.path.join(opt.checkpoints_dir, opt.name))
        
    return mod_id_seg, input_id


import copy
def map_model_names(model_names,mod_id_seg_source,input_id_source,mod_id_seg_target,input_id_target):
    """
    Used for DeepLIIFKD to map model/image names from teacher model to student model,
    when mod_id_seg and/or input_id is different.
    """
    d_res = {}
    for model_name in model_names:
        model_name_new = copy.deepcopy(model_name)
        if len(model_name) > 2 and model_name[1] == mod_id_seg_source:
            model_name_new = model_name[0]+mod_id_seg_target+model_name[2:] # replace mod id seg
            if input_id_source != input_id_target: # currently only support 0 or 1 as input_id
                if int(input_id_target) == 0:
                    model_name_new = model_name_new[:2] + str(int(model_name_new[2:])-1)
                else:
                    model_name_new = model_name_new[:2] + str(int(model_name_new[2:])+1)
        d_res[model_name] = model_name_new
    # this is not a model name but the final aggregated segmentation image name
    # so it cannot be obtained from model names
    d_res['G'+mod_id_seg_source] = 'G'+mod_id_seg_target
    return d_res



import math
import warnings

import numpy as np
from numba import jit, typed
from PIL import Image
Image.MAX_IMAGE_PIXELS = None


def imadjust(x, gamma=0.7, c=0, d=1):
    """
    Adjusting the image contrast and brightness

    :param x: Input array
    :param gamma: Gamma value
    :param c: Minimum value
    :param d: Maximum value
    :return: Adjusted image
    """
    a = x.min()
    b = x.max()
    y = (((x - a) / (b - a)) ** gamma) * (d - c) + c
    return y


def adjust_dapi(inferred_tile, orig_tile):
    """Adjusts the intensity of mpIF DAPI image

    Parameters:
        inferred_tile (Image) -- inferred tile image
        orig_tile (Image) -- original tile image

    Returns:
        processed_tile (Image) -- the adjusted mpIF DAPI image
    """
    inferred_tile_array = np.array(inferred_tile)
    orig_tile_array = np.array(orig_tile)

    multiplier = 8 / math.log(np.max(orig_tile_array))

    if np.mean(orig_tile_array) < 200:
        processed_tile = imadjust(inferred_tile_array,
                                  gamma=multiplier * math.log(np.mean(inferred_tile_array)) / math.log(
                                      np.mean(orig_tile_array)),
                                  c=5, d=255).astype(np.uint8)

    else:
        processed_tile = imadjust(inferred_tile_array,
                                  gamma=multiplier,
                                  c=5, d=255).astype(np.uint8)
    return Image.fromarray(processed_tile)


def adjust_marker(inferred_tile, orig_tile):
    """Adjusts the intensity of mpIF marker image

    Parameters:
        inferred_tile (Image) -- inferred tile image
        orig_tile (Image) -- original tile image

    Returns:
        processed_tile (Image) -- the adjusted marker image
    """
    inferred_tile_array = np.array(inferred_tile)
    orig_tile_array = np.array(orig_tile)

    multiplier = 8 / math.log(np.max(orig_tile_array))

    if np.mean(orig_tile_array) < 200:
        processed_tile = imadjust(inferred_tile_array,
                                  gamma=multiplier * math.log(np.std(inferred_tile_array)) / math.log(
                                      np.std(orig_tile_array)),
                                  c=5, d=255).astype(np.uint8)

    else:
        processed_tile = imadjust(inferred_tile_array,
                                  gamma=multiplier,
                                  c=5, d=255).astype(np.uint8)
    return Image.fromarray(processed_tile)


# Default postprocessing values
DEFAULT_SEG_THRESH = 120
DEFAULT_NOISE_THRESH = 4

# Values for uint8 label masks
LABEL_UNKNOWN = 50
LABEL_POSITIVE = 200
LABEL_NEGATIVE = 150
LABEL_BACKGROUND = 0
LABEL_CELL = 100
LABEL_BORDER_POS = 220
LABEL_BORDER_NEG = 170
LABEL_BORDER_POS2 = 221
LABEL_BORDER_NEG2 = 171


def to_array(img, grayscale=False):
    """
    Convert a color image to an array of pixels.

    Parameters
    ----------
    img : Image | ndarray
        Image to convert. If an array is provided instead, it is used directly.
    grayscale : bool
        Whether the input image should be converted to grayscale
        by taking the maximum channel value for each pixel.

    Returns
    -------
    ndarray
        A 2D (grayscale) or 3D array with the pixels of the converted image.
    """

    if isinstance(img, Image.Image):
        img = np.asarray(img) if img.mode == 'RGB' else np.asarray(img.convert('RGB'))
    if grayscale and len(img.shape) == 3:
        img = img.max(axis=-1)
    return img


@jit(nopython=True)
def in_bounds(array, index):
    """
    Check if an index is valid for an array.

    Parameters
    ----------
    array : ndarray
        2D array.
    index : tuple
        2-element tuple with index values matching the array shape (e.g., for a pixel array where
        array.shape[0] is height and array.shape[1] is width, then index will be in (y, x) order).

    Returns
    -------
    bool
        Whether or not the index is within the bounds of the array.
    """

    return index[0] >= 0 and index[0] < array.shape[0] and index[1] >= 0 and index[1] < array.shape[1]


@jit(nopython=True)
def create_posneg_mask(seg, thresh):
    """
    Create a mask of positive and negative pixels from the segmentation image.

    Parameters
    ----------
    seg : ndarray
        3D uint8 array (2D image w/ 3 channels) with segmentation probabilities.
    thresh : int
        Threshold to use in determining if a pixel should be labeled as positive/negative.

    Returns
    -------
    ndarray
        2D uint8 array of mask values with every pixel labeled as unknown, positive, or negative.
    """

    mask = np.full(seg.shape[0:2], LABEL_UNKNOWN, dtype=np.uint8)
    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):
            if seg[y, x, 0] + seg[y, x, 2] > thresh and seg[y, x, 1] <= 80:
                if seg[y, x, 0] >= seg[y, x, 2]:
                    mask[y, x] = LABEL_POSITIVE
                else:
                    mask[y, x] = LABEL_NEGATIVE

    return mask


@jit(nopython=True)
def mark_background(mask):
    """
    Mask all background pixels in-place by 4-connected region growing unknown boundary pixels.

    Parameters
    ----------
    mask: ndarray
        2D uint8 array with pixels labeled as positive, negative, or unknown.
        After the function executes, the pixels will be labeled as background, positive, negative, or unknown.
    """

    for i in range(mask.shape[0]):
        if mask[i, 0] == LABEL_UNKNOWN:
            mask[i, 0] = LABEL_BACKGROUND
        if mask[i, mask.shape[1]-1] == LABEL_UNKNOWN:
            mask[i, mask.shape[1]-1] = LABEL_BACKGROUND
    for j in range(mask.shape[1]):
        if mask[0, j] == LABEL_UNKNOWN:
            mask[0, j] = LABEL_BACKGROUND
        if mask[mask.shape[0]-1, j] == LABEL_UNKNOWN:
            mask[mask.shape[0]-1, j] = LABEL_BACKGROUND

    count = 1
    while count > 0:
        count = 0
        for i in range(mask.shape[0]):
            for j in range(mask.shape[1]):
                if mask[i, j] == LABEL_UNKNOWN:
                    if (mask[i-1, j] == LABEL_BACKGROUND or mask[i+1, j] == LABEL_BACKGROUND or
                            mask[i, j-1] == LABEL_BACKGROUND or mask[i, j+1] == LABEL_BACKGROUND):
                        mask[i, j] = LABEL_BACKGROUND
                        count += 1
        if count > 0:
            for i in range(mask.shape[0]-1, -1, -1):
                for j in range(mask.shape[1]-1, -1, -1):
                    if mask[i, j] == LABEL_UNKNOWN:
                        if (mask[i-1, j] == LABEL_BACKGROUND or mask[i+1, j] == LABEL_BACKGROUND or
                                mask[i, j-1] == LABEL_BACKGROUND or mask[i, j+1] == LABEL_BACKGROUND):
                            mask[i, j] = LABEL_BACKGROUND


@jit(nopython=True)
def compute_cell_mapping(mask, marker, noise_thresh, large_noise_thresh):
    """
    Compute the mapping from mask to positive and negative cells.

    Parameters
    ----------
    mask : ndarray
        2D uint8 array with pixels labeled as positive, negative, background, or unknown.
        After the function executes, the pixels will be labeled as background or cell.
    marker : ndarray
        2D uint8 array with the inferred marker values.

    Returns
    -------
    typed.List[tuple] :
        Cell data as a list of 7-element tuples with the following:
        [0] - number of pixels in the cell
        [1] - whether the cell is positive (True) or negative (False)
        [2] - marker value for the cell
        [3-4] - first pixel coordinates (x, y) of the cell cluster
        [5-6] - centroid of the cell (x, y)
    """

    neighbors = [(-1, -1), (0, -1), (1, -1), (-1, 0), (1, 0), (-1, 1), (0, 1), (1, 1)]
    cells = typed.List()

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):
            if mask[y, x] != LABEL_BACKGROUND and mask[y, x] != LABEL_CELL:
                seeds = [(y, x)]
                count = 1
                count_positive = 1 if mask[y, x] == LABEL_POSITIVE else 0
                count_negative = 1 if mask[y, x] == LABEL_NEGATIVE else 0
                max_marker = marker[y, x] if marker is not None else 0
                mask[y, x] = LABEL_CELL
                center_y = y
                center_x = x

                while len(seeds) > 0:
                    seed = seeds.pop()
                    for n in neighbors:
                        idx = (seed[0] + n[0], seed[1] + n[1])
                        if in_bounds(mask, idx) and mask[idx] != LABEL_BACKGROUND and mask[idx] != LABEL_CELL:
                            seeds.append(idx)
                            if mask[idx] == LABEL_POSITIVE:
                                count_positive += 1
                            elif mask[idx] == LABEL_NEGATIVE:
                                count_negative += 1
                            if marker is not None and marker[idx] > max_marker:
                                max_marker = marker[idx]
                            mask[idx] = LABEL_CELL
                            center_y += idx[0]
                            center_x += idx[1]
                            count += 1

                if count > noise_thresh and (large_noise_thresh is None or count < large_noise_thresh):
                    center_y = int(round(center_y / count))
                    center_x = int(round(center_x / count))
                    positive = True if count_positive >= count_negative else False
                    cells.append((count, positive, max_marker, x, y, center_x, center_y))

    return cells


def get_cells_info(seg, marker, resolution, noise_thresh, seg_thresh, large_noise_thresh):
    """
    Find all cells in the segmentation image that are larger than the noise threshold.

    Parameters
    ----------
    seg : Image | ndarray
        Inferred segmentation map image.
    marker : Image | ndarray
        Inferred marker image.
    resolution: string
        The resolution/magnification of the original image.  Valid values are '10x', '20x', or '40x'.
    noise_thresh : int
        Threshold for tiny noise to ignore (include only cells larger than this value).
    seg_thresh : int
        Threshold to use in determining if a pixel should be labeled as positive/negative.
    large_noise_thresh : int | None
        Threshold for large noise to ignore (include only cells smaller than this value).

    Returns
    -------
    ndarray :
        Label mask.
    typed.List[tuple] :
        Cell data as a list of 7-element tuples.
    dict :
        Calculated default values.
    """

    seg = to_array(seg)
    if marker is not None:
        marker = to_array(marker, True)

    mask = create_posneg_mask(seg, seg_thresh)
    mark_background(mask)
    cellsinfo = compute_cell_mapping(mask, marker, noise_thresh, large_noise_thresh)

    defaults = {}
    sizes = np.zeros(len(cellsinfo), dtype=np.int64)
    for i in range(len(cellsinfo)):
        sizes[i] = cellsinfo[i][0]
    defaults['size_thresh'] = calculate_default_size_threshold(sizes, resolution)
    if marker is not None:
        defaults['marker_thresh'] = calculate_default_marker_threshold(marker)

    return mask, cellsinfo, defaults


@jit(nopython=True)
def create_kde(values, count, bandwidth = 1.0):
    """
    Create Gaussian kernel density estimate (KDE) for values with count number of bins.

    Parameters
    ----------
    values : list
        Input values.
    count : int
        Number of bins for KDE.
    bandwidth: float
        Bandwidth (smoothing parameter) for KDE.

    Returns
    -------
    ndarray :
        Kernel density estimate.
    float :
        Step size.
    """

    gaussian_denom_inv = 1 / math.sqrt(2 * math.pi);
    max_value = max(values) + 1;
    step = max_value / count;
    n = values.shape[0];
    h = bandwidth;
    h_inv = 1 / h;
    kde = np.zeros(count, dtype=np.float32)

    for i in range(count):
        x = i * step
        total = 0
        for j in range(n):
            val = (x - values[j]) * h_inv;
            total += math.exp(-(val*val/2)) * gaussian_denom_inv; # Gaussian
        kde[i] = total / (n*h);

    return kde, step


@jit(nopython=True)
def calculate_default_size_threshold(cell_sizes, resolution='40x'):
    """
    Calculate a default size threshold to exclude small cells.

    Parameters
    ----------
    cell_sizes : ndarray
        1D array of cell sizes.
    resolution : string
        The resolution/magnification of the original image.  Valid values are '10x', '20x', or '40x'.

    Returns
    -------
    int :
        Default size threshold.
    """

    if cell_sizes.shape[0] > 1:
        kde, step = create_kde(np.sqrt(cell_sizes), 500)
        idx = 1
        for i in range(1, kde.shape[0]-1):
            if kde[i] < kde[i-1] and kde[i] < kde[i+1]:
                idx = i
                break
        thresh_sqrt = (idx - 1) * step

        allowed_range_sqrt = (4, 7, 10) # [min, default, max] for default sqrt size thresh at 40x
        if resolution == '20x':
            allowed_range_sqrt = (3, 4, 6)
        elif resolution == '10x':
            allowed_range_sqrt = (2, 2, 3)

        if thresh_sqrt < allowed_range_sqrt[0]:
            thresh_sqrt = allowed_range_sqrt[0]
        elif thresh_sqrt > allowed_range_sqrt[2]:
            thresh_sqrt = allowed_range_sqrt[1]

        return round(thresh_sqrt * thresh_sqrt)

    else:
        return 0


def calculate_stain_range(stain):
    """
    Calculate the range of the 99.9 percentile of non-zero pixels in the stain image.

    Parameters
    ----------
    stain : ndarray
        2D uint8 array (image).

    Returns
    -------
    tuple[int] :
        2-element tuple with 0.1 and 99.9 percentile values.
    """

    nonzero = stain[stain != 0]
    if nonzero.shape[0] > 0:
        return (round(np.percentile(nonzero, 0.1)), round(np.percentile(nonzero, 99.9)))
    else:
        return (0, 0)


def calculate_default_marker_threshold(marker):
    """
    Calculate a default threshold for a marker image as 90% of the 99.9 percentile range.

    Parameters
    ----------
    marker : ndarray
        2D uint8 array (image).

    Results
    -------
    int :
        Default marker threshold.
    """

    marker_range = calculate_stain_range(marker)
    return round((marker_range[1] - marker_range[0]) * 0.9) + marker_range[0]


@jit(nopython=True)
def get_cell_boundary(mask, x, y):
    """
    Get the boundary contour pixels for a cell, and also the bounding box.
    The provided starting (x, y) pixel must be the first pixel encountered
    from the top left, whether found by searching via rows or columns.

    Parameters
    ----------
    mask : ndarray
        2D uint8 ndarray of background and cell labels
    x : int
        x-coordinate of the first pixel of the cell
    y : int
        y-coordinate of the first pixel of the cell

    Returns
    -------
    list :
        Bounding box of the cell as a list of two 2-element tuples.
    list :
        All boundary pixels (x, y) going clockwise from first point.
    """

    w = mask.shape[1]
    h = mask.shape[0]

    if not in_bounds(mask, (y, x)) or mask[y, x] == LABEL_BACKGROUND:
        return None, None

    '''
    In normal xy coordinates, check neighbors clockwise in the following order:
     0 1 2
     7 - 3
     6 5 4
    List neighbors in xy coordinates, but xy are switched to yx for numpy array access.
    '''
    neighbors = [(-1, -1), (0, -1), (1, -1), (1, 0), (1, 1), (0, 1), (-1, 1), (-1, 0)] # (dx, dy)
    neighbors *= 2

    boundary = [(x, y)]
    min_x = x
    min_y = y
    max_x = x
    max_y = y

    # Go counter-clockwise to find previous pixel
    idx = 6
    while idx >= 0:
        nx = x + neighbors[idx][0]
        ny = y + neighbors[idx][1]
        if in_bounds(mask, (ny, nx)) and mask[ny, nx] != LABEL_BACKGROUND:
            break
        idx -= 1
    if idx < 0:
        return [(x, y), (x, y)], [(x, y)]

    px = x + neighbors[idx][0]
    py = y + neighbors[idx][1]
    boundary = [(px, py), (x, y)]

    # Go clockwise to get border pixels in order
    while True:
        dx = px - x
        dy = py - y
        idx = neighbors.index((dx, dy)) + 1
        while True:
            nx = x + neighbors[idx][0]
            ny = y + neighbors[idx][1]
            if in_bounds(mask, (ny, nx)) and mask[ny, nx] != LABEL_BACKGROUND:
                break
            idx += 1
        px = x
        py = y
        x = nx
        y = ny
        boundary.append((x, y))

        if x < min_x:
            min_x = x
        elif x > max_x:
            max_x = x
        if y < min_y:
            min_y = y
        elif y > max_y:
            max_y = y

        if px == boundary[0][0] and py == boundary[0][1] and x == boundary[1][0] and y == boundary[1][1]:
            break

    return [(min_x, min_y), (max_x, max_y)], boundary[1:-1]


def make_simple_contour(points):
    """
    Make a simplified version of a contour by removing redundant points within
    straight lines (i.e., each straight line segment will be reduced to contain
    only the first and last point for that segment).  It is assumed that the
    vectors between points are all one of the eight pixel neighbor directions.
    This means that either one of the x- or y-direction must be zero, or if
    both values are non-zero then they must be the same (i.e., [1,0], [0,2],
    [2,2], and [3,3] are vall valid direction vectors, but [1,2] is not).
    The input parameter of contour points must contain at least one point.

    Parameters
    ----------
    points : list
        Contour of boundary points (x, y).

    Returns
    -------
    list :
        Simplified contour of boundary points (x, y).
    """

    # always keep first point
    simple = [(points[0][0], points[0][1])]

    # if only one point in contour, then done
    if len(points) == 1:
        return simple

    # for all middle points (exclude first and last)
    for i in range(1, len(points) - 1):
        dx0 = points[i][0] - points[i-1][0]
        dy0 = points[i][1] - points[i-1][1]
        dx1 = points[i+1][0] - points[i][0]
        dy1 = points[i+1][1] - points[i][1]
        same_dx = (dx0 == dx1) or (dx0 > 0 and dx1 > 0) or (dx0 < 0 and dx1 < 0)
        same_dy = (dy0 == dy1) or (dy0 > 0 and dy1 > 0) or (dy0 < 0 and dy1 < 0)
        if not same_dx or not same_dy:
            simple.append((points[i][0], points[i][1]))

    # for last point (calculate p[n]-p[n-1] and p[0]-p[n])
    dx0 = points[-1][0] - points[-2][0]
    dy0 = points[-1][1] - points[-2][1]
    dx1 = points[0][0] - points[-1][0]
    dy1 = points[0][1] - points[-1][1]
    same_dx = (dx0 == dx1) or (dx0 > 0 and dx1 > 0) or (dx0 < 0 and dx1 < 0)
    same_dy = (dy0 == dy1) or (dy0 > 0 and dy1 > 0) or (dy0 < 0 and dy1 < 0)
    if not same_dx or not same_dy:
        simple.append((points[-1][0], points[-1][1]))

    return simple


def make_full_contour(points):
    """
    Convert a simplified contour to a complete pixel-by-pixel contour
    (i.e., every point is an 8-neighbor of the previous point).  It is
    assumed that vectors between points are all one of the eight pixel
    neighbor directions.  The input parameter of contour points must
    contain at least one point.

    Parameters
    ----------
    points : list
        Contour of boundary points (x, y).

    Returns
    -------
    list :
        Full contour of boundary points (x, y).
    """

    # start with first point
    full = [(points[0][0], points[0][1])]

    # for all remaining points:
    for i in range(1, len(points)):

        # calculate direction from last full point to current input point
        dx = points[i][0] - full[-1][0]
        dy = points[i][1] - full[-1][1]
        dx = 1 if dx > 0 else (-1 if dx < 0 else 0)
        dy = 1 if dy > 0 else (-1 if dy < 0 else 0)

        # add direction to last full point until reach current input point
        while full[-1][0] != points[i][0] or full[-1][1] != points[i][1]:
            full.append((full[-1][0] + dx, full[-1][1] + dy))

    # calculate direction from last full point until first point
    dx = full[0][0] - full[-1][0]
    dy = full[0][1] - full[-1][1]
    dx = 1 if dx > 0 else (-1 if dx < 0 else 0)
    dy = 1 if dy > 0 else (-1 if dy < 0 else 0)

    # add direction to last full point until reach first point (avoid duplicate)
    while full[-1][0] + dx != full[0][0] or full[-1][1] + dy != full[0][1]:
        full.append((full[-1][0] + dx, full[-1][1] + dy))

    return full


def to_base92(values, min_len=1):
    """
    Convert integer values to base92, offset by 35 for printable ASCII values
    (i.e., output characters for [0-91] are in the range [35-126]).
    All encodings will have the same number of characters, of at least min_len
    (i.e., smaller in length encodings will be padded with 35).

    Parameters
    ----------
    values : int | list[int] | tuple[int]
        The integer value(s) to a base92 ASCII encoding.
    min_len : int
        The minimum number of characters for the base92 ASCII encoding of each value.

    Returns
    -------
    string | list[string] :
        The converted value(s).
    """

    multi = type(values) is list or type(values) is tuple
    if not multi:
        values = [values]

    results = []
    for val in values:
        res = ''
        while val > 0:
            res += chr((val % 92) + 35)
            val //= 92
        results.append(res)

    max_len = max(len(r) for r in results)
    fixed_len = max_len if max_len > min_len else min_len

    for i in range(len(results)):
        while len(results[i]) < fixed_len:
            results[i] += chr(35)
        results[i] = results[i][::-1]

    if not multi:
        results = results[0]
    return results


def from_base92(val):
    """
    Convert from base92 ASCII encoding to integer value.

    Parameters
    ----------
    val : string
        The base92 ASCII encoded value.

    Returns
    -------
    int :
        The converted value.
    """

    res = 0
    for v in val:
        res *= 92
        res += (ord(v) - 35)
    return res


def encode_cell_data_v4(data):
    """
    Encode as v4 the provided cell data to string.

    Parameters
    ----------
    data : dict
        Dictionary of cell data.

    Returns
    -------
    string :
        Encoded cell data as a single ASCII string.
    """

    cell = '' # encoded cell data as string

    # encode cell size (in pixels)
    size = to_base92(data['size'])
    size_len = len(size)
    cell += size

    # encode cell classification (pos/neg) and marker value
    positive = int(data['positive'])
    marker = data['marker']
    classification = (marker * 2) + positive
    cell += to_base92(classification, 2)

    # encode anchor point (bbox top left) and extent (bbox bottom right)
    topleft = to_base92(data['bbox'][0])
    topleft_len = len(topleft[0])
    cell += topleft[0]
    cell += topleft[1]

    # encode extent (bbox bottom right), centroid, and first boundary contour point
    # as offsets from the previously encoded anchor point (bbox top left)
    x = data['bbox'][0][0]
    y = data['bbox'][0][1]
    offsets = [*data['bbox'][1], *data['centroid'], *data['boundary'][0]]
    for j in range(0, len(offsets), 2):
        offsets[j] -= x
        offsets[j+1] -= y
    offsets = to_base92(offsets)
    offsets_len = len(offsets[0])
    cell += offsets[0]
    cell += offsets[1]
    cell += offsets[2]
    cell += offsets[3]
    cell += offsets[4]
    cell += offsets[5]

    # encode number of chars for variable length encodations and prepend to cell string
    encoded_lens = ((size_len - 1) * 16) + ((topleft_len - 1) * 4) + (offsets_len - 1)
    encoded_lens = chr(encoded_lens + 35)
    cell = encoded_lens + cell

    # encode remaining boundary contour points using Freeman chain code
    # Freeman chain code:
    #   3  2  1
    #    \ | /
    #   4-- --0
    #    / | \
    #   5  6  7
    boundary = ''
    for j in range(1, len(data['boundary'])):
        dx = data['boundary'][j][0] - data['boundary'][j-1][0]
        dy = data['boundary'][j][1] - data['boundary'][j-1][1]
        if dx >= 1 and dy == 0:
            direction = 0
        elif dx >= 1 and dy <= -1:
            direction = 1
        elif dx == 0 and dy <= -1:
            direction = 2
        elif dx <= -1 and dy <= -1:
            direction = 3
        elif dx <= -1 and dy == 0:
            direction = 4
        elif dx <= -1 and dy >= 1:
            direction = 5
        elif dx == 0 and dy >= 1:
            direction = 6
        elif dx >= 1 and dy >= 1:
            direction = 7
        else: # this should not (cannot) happen, so if it does, then exit
            exit()
        distance = max(abs(dx), abs(dy))
        if distance == 0: # this should not (cannot) happen, but if duplicate point, then skip
            continue
        while distance > 10:
            encoded = (10 * 8) + direction
            boundary += chr(encoded + 35)
            distance -= 10
        encoded = (distance * 8) + direction
        boundary += chr(encoded + 35)
    cell += boundary

    return cell


def decode_cell_data_v4(cell):
    """
    Decode v4 encoded cell string and return dictionary of cell data.

    Parameters
    ----------
    cell : string
        Encoded cell data as a single ASCII string.

    Returns
    -------
    dict :
        Dictionary with the decoded cell data.
    """

    data = {} # decoded cell data

    # decode number of chars for variable length encodations
    n = ord(cell[0]) - 35
    ns = (n // 16) + 1      # num chars for cell size
    na = ((n // 4) % 4) + 1 # num chars for anchor coordinates
    no = (n % 4) + 1        # num chars for offset coordinates

    # decode cell size (in pixels)
    data['size'] = from_base92(cell[1:1+ns])

    # decode cell classification (pos/neg) and marker value
    classification = from_base92(cell[1+ns:3+ns])
    data['positive'] = bool(classification % 2)
    data['marker'] = classification // 2

    # decode anchor point (bbox top left) and extent (bbox bottom right)
    x = from_base92(cell[3+ns:3+ns+na])
    y = from_base92(cell[3+ns+na:3+ns+2*na])
    ex = x + from_base92(cell[3+ns+2*na:3+ns+2*na+no])
    ey = y + from_base92(cell[3+ns+2*na+no:3+ns+2*na+2*no])
    data['bbox'] = [(x, y), (ex, ey)]

    # decode centroid point
    cx = x + from_base92(cell[3+ns+2*na+2*no:3+ns+2*na+3*no])
    cy = y + from_base92(cell[3+ns+2*na+3*no:3+ns+2*na+4*no])
    data['centroid'] = (cx, cy)

    # decode first boundary contour points
    bx = x + from_base92(cell[3+ns+2*na+4*no:3+ns+2*na+5*no])
    by = y + from_base92(cell[3+ns+2*na+5*no:3+ns+2*na+6*no])
    data['boundary'] = [(bx, by)]

    # directions using Freeman chain code
    freeman = [(1,0), (1,-1), (0,-1), (-1,-1), (-1,0), (-1,1), (0,1), (1,1)]

    # decode remaining boundary contour points
    prev_direction = None
    for c in cell[3+ns+2*na+6*no:]:
        point = ord(c) - 35
        distance = point // 8
        direction = freeman[point % 8]
        movement = (direction[0]*distance, direction[1]*distance)
        px = data['boundary'][-1][0] + movement[0]
        py = data['boundary'][-1][1] + movement[1]
        if direction == prev_direction:
            data['boundary'].pop()
        data['boundary'].append((px, py))
        prev_direction = direction

    return data


@jit(nopython=True)
def create_cell_classification(mask, cellsinfo,
                               size_thresh=0,
                               marker_thresh=None,
                               size_thresh_upper=None):
    """
    Create final cell classification in-place for the mask and
    calculate counts for positive and negative cell counts.

    Parameters
    ----------
    mask : ndarray
        2D uint8 array label map.
    cellsinfo : list
        Information about each cell found from the segmentation.
    size_thresh : int
        Include only cells larger than this size.
    marker_thresh : int
        Make cell positive if marker value is above this threshold (override original classification).
    size_thresh_upper : int
        Include only cells smaller than this size.

    Results
    -------
    dict :
        Dictionary with the counts of positive, negative, and total cells.
    """

    neighbors = [(-1, -1), (0, -1), (1, -1), (-1, 0), (1, 0), (-1, 1), (0, 1), (1, 1)]
    border_neighbors = [(0, -1), (-1, 0), (1, 0), (0, 1)]
    num_pos, num_neg = 0, 0
    if marker_thresh is None:
        marker_thresh = 255

    for cell in cellsinfo:
        if cell[0] > size_thresh and (size_thresh_upper is None or cell[0] < size_thresh_upper):
            is_pos = True if cell[1] or cell[2] > marker_thresh else False
            if is_pos:
                label = LABEL_POSITIVE
                label_border = LABEL_BORDER_POS
                num_pos += 1
            else:
                label = LABEL_NEGATIVE
                label_border = LABEL_BORDER_NEG
                num_neg += 1

            x = cell[3]
            y = cell[4]
            mask[y,x] = label_border
            seeds = [(y, x)]

            while len(seeds) > 0:
                seed = seeds.pop()
                for n in neighbors:
                    idx = (seed[0] + n[0], seed[1] + n[1])
                    if in_bounds(mask, idx) and mask[idx] == LABEL_CELL:
                        seeds.append(idx)
                        for n in border_neighbors:
                            idx2 = (idx[0] + n[0], idx[1] + n[1])
                            if in_bounds(mask, idx2) and mask[idx2] == LABEL_BACKGROUND:
                                mask[idx2] = label_border
                        mask[idx] = label

    num_total = num_pos + num_neg
    return {
        'num_total': num_total,
        'num_pos': num_pos,
        'num_neg': num_neg,
    }


@jit(nopython=True)
def enlarge_cell_boundaries(mask):
    """
    Enlarge cell boundaries in-place in mask by one pixel in each direction.

    Parameters
    ----------
    mask : ndarray
        2D uint8 label map.
    """

    neighbors = [(-1, -1), (0, -1), (1, -1), (-1, 0), (1, 0), (-1, 1), (0, 1), (1, 1)]

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):
            if mask[y, x] == LABEL_BORDER_POS or mask[y, x] == LABEL_BORDER_NEG:
                value = LABEL_BORDER_POS2 if mask[y, x] == LABEL_BORDER_POS else LABEL_BORDER_NEG2
                for n in neighbors:
                    idx = (y + n[0], x + n[1])
                    if in_bounds(mask, idx) and mask[idx] == LABEL_BACKGROUND:
                        mask[idx] = value

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):
            if mask[y, x] == LABEL_BORDER_POS2:
                mask[y, x] = LABEL_BORDER_POS
            elif mask[y, x] == LABEL_BORDER_NEG2:
                mask[y, x] = LABEL_BORDER_NEG


@jit(nopython=True)
def create_final_images(overlay, mask):
    """
    Create the final overlay (in-place) and refined images from the mask.
    The 'overlay' parameter is the image on which to create the overlay,
    which will be done in-place.

    Parameters
    ----------
    overlay : ndarray
        3D uint8 array (2D image of 3 channels).
        Generally, a copy of the original input image.
    mask : ndarray
        2D uint8 label map.

    Returns
    -------
    ndarray :
        3D uint8 array (2D image of 3 channels) containing the overlaid image.
    ndarray :
        3D uint8 array (2D image of 3 channels) containing the refined segmentation image.
    """

    refined = np.zeros_like(overlay)

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):
            if mask[y, x] == LABEL_BORDER_POS:
                overlay[y, x] = (255, 0, 0)
                refined[y, x, 1] = 255
            elif mask[y, x] == LABEL_BORDER_NEG:
                overlay[y, x] = (0, 0, 255)
                refined[y, x, 1] = 255
            elif mask[y, x] == LABEL_POSITIVE:
                refined[y, x, 0] = 255
            elif mask[y, x] == LABEL_NEGATIVE:
                refined[y, x, 2] = 255

    return overlay, refined


@jit(nopython=True)
def fill_cells(mask):
    """
    For a mask with cell outlines, fill in the center of the cells in-place.
    The cell outlines must surround cell entirely, including image border pixels.

    Parameters
    ----------
    mask : ndarray
        2D uint8 label map.
    """

    for y in range(mask.shape[0]):
        for x in range(1, mask.shape[1]):
            if mask[y, x] == LABEL_UNKNOWN:
                if mask[y, x-1] == LABEL_BORDER_POS or mask[y, x-1] == LABEL_POSITIVE:
                    mask[y, x] = LABEL_POSITIVE
                else:
                    mask[y, x] = LABEL_NEGATIVE

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):
            if mask[y, x] == LABEL_BORDER_POS:
                mask[y, x] = LABEL_POSITIVE
            elif mask[y, x] == LABEL_BORDER_NEG:
                mask[y, x] = LABEL_NEGATIVE


@jit(nopython=True)
def create_outer_boundary(mask):
    """
    For a mask with positive and negative cell labels,
    set pixels neighboring the cells in-place to border labels.

    Parameters
    ----------
    mask : ndarray
        2D uint8 label map.
    """

    border_neighbors = [(0, -1), (-1, 0), (1, 0), (0, 1)]

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):
            if mask[y, x] == LABEL_POSITIVE or mask[y, x] == LABEL_NEGATIVE:
                for n in border_neighbors:
                    idx2 = (y + n[0], x + n[1])
                    if in_bounds(mask, idx2) and mask[idx2] == LABEL_BACKGROUND:
                        mask[idx2] = LABEL_BORDER_POS if mask[y, x] == LABEL_POSITIVE else LABEL_BORDER_NEG


def calculate_large_noise_thresh(large_noise_thresh, resolution):
    if large_noise_thresh != 'default':
        return large_noise_thresh
    if resolution == '10x':
        return 1000
    elif resolution == '20x':
        return 4000
    else: # 40x
        return 16000


def compute_cell_results(seg, marker, resolution, version=3,
                         seg_thresh=DEFAULT_SEG_THRESH,
                         noise_thresh=DEFAULT_NOISE_THRESH,
                         large_noise_thresh=None):
    """
    Perform postprocessing to compute individual cell results.

    Parameters
    ----------
    seg : Image | ndarray
        Inferred segmentation map image.
    marker : Image | ndarray
        Inferred marker image.
    resolution : string
        The resolution/magnification of the original image.  Valid values are '10x', '20x', or '40x'.
    version : int
        Version of the cell data (valid values are 3 and 4).
    seg_thresh : int
        Threshold to use in determining if a pixel should be labeled as positive/negative.
    noise_thresh : int
        Threshold for tiny noise to ignore (include only cells larger than this value).
    large_noise_thresh : int | string | None
        Threshold for large noise to ignore (include only cells smaller than this value).
        Valid arguments can be an integer value, the string value 'default', or None.

    Returns
    -------
    dict :
        Individual cell data and other associated values.
    """

    if version not in [3, 4]:
        warnings.warn('Invalid cell data version provided, defaulting to version 3.')
        version = 3

    large_noise_thresh = calculate_large_noise_thresh(large_noise_thresh, resolution)
    mask, cellsinfo, defaults = get_cells_info(seg, marker, resolution, noise_thresh, seg_thresh, large_noise_thresh)

    cells = []
    for cell in cellsinfo:
        bbox, boundary = get_cell_boundary(mask, cell[3], cell[4])
        data = {
            'size': cell[0],
            'positive': cell[1],
            'marker': cell[2],
            'bbox': bbox,
            'centroid': (cell[5], cell[6]),
            'boundary': make_simple_contour(boundary),
        }
        if version == 4:
            data = encode_cell_data_v4(data)
        cells.append(data)

    results = {
        'cells': cells,
        'settings': {
            'default_marker_thresh': defaults['marker_thresh'] if 'marker_thresh' in defaults else None,
            'default_size_thresh': defaults['size_thresh'],
            'noise_thresh': noise_thresh,
            'large_noise_thresh': large_noise_thresh,
            'seg_thresh': seg_thresh,
        },
        'dataVersion': version,
    }

    return results


def compute_final_results(orig, seg, marker, resolution,
                          size_thresh='default',
                          marker_thresh=None,
                          size_thresh_upper=None,
                          seg_thresh=DEFAULT_SEG_THRESH,
                          noise_thresh=DEFAULT_NOISE_THRESH,
                          large_noise_thresh=None):
    """
    Perform postprocessing to compute final count and image results.

    Parameters
    ----------
    orig : Image | ndarray
        Original input image.
    seg : Image | ndarray
        Inferred segmentation map image.
    marker : Image | ndarray
        Inferred marker image.
    resolution : string
        The resolution/magnification of the original image.  Valid values are '10x', '20x', or '40x'.
    size_thresh : int
        Include only cells larger than this size.
    marker_thresh : int
        Make cell positive if marker value is above this threshold (override original classification).
    size_thresh_upper : int
        Include only cells smaller than this size.
    seg_thresh : int
        Threshold to use in determining if a pixel should be labeled as positive/negative.
    noise_thresh : int
        Threshold for tiny noise to ignore (include only cells larger than this value).
    large_noise_thresh : int | string | None
        Threshold for large noise to ignore (include only cells smaller than this value).
        Valid arguments can be an integer value, the string value 'default', or None.

    Returns
    -------
    ndarray :
        3D uint8 array (2D image of 3 channels) containing the overlaid image.
    ndarray :
        3D uint8 array (2D image of 3 channels) containing the refined segmentation image.
    dict :
        Dictionary with scoring and settings information.
    """

    large_noise_thresh = calculate_large_noise_thresh(large_noise_thresh, resolution)
    mask, cellsinfo, defaults = get_cells_info(seg, marker, resolution, noise_thresh, seg_thresh, large_noise_thresh)

    if size_thresh is None:
        size_thresh = 0
    elif size_thresh == 'default':
        size_thresh = defaults['size_thresh']
    if marker_thresh == 'default':
        marker_thresh = defaults['marker_thresh']

    counts = create_cell_classification(mask, cellsinfo, size_thresh, marker_thresh, size_thresh_upper)
    enlarge_cell_boundaries(mask)
    enlarge_cell_boundaries(mask)
    overlay, refined = create_final_images(np.array(orig), mask)

    scoring = {
        'num_total': counts['num_total'],
        'num_pos': counts['num_pos'],
        'num_neg': counts['num_neg'],
        'percent_pos': round(counts['num_pos'] / counts['num_total'] * 100, 1) if counts['num_pos'] > 0 else 0,
        'seg_thresh': seg_thresh,
        'size_thresh': size_thresh,
        'size_thresh_upper': size_thresh_upper,
        'marker_thresh': marker_thresh if marker is not None else None,
    }

    return overlay, refined, scoring


def cells_to_final_results(data, orig,
                           size_thresh='default',
                           marker_thresh=None,
                           size_thresh_upper=None):
    """
    Compute final count and image results from previously postprocessed cell results.

    Parameters
    ----------
    data : dict
        Individual cell data and associated values generated by the 'compute_cell_results' function.
    orig : Image | ndarray
        Original input image.
    size_thresh : int
        Include only cells larger than this size.
    marker_thresh : int
        Make cell positive if marker value is above this threshold (override original classification).
    size_thresh_upper : int
        Include only cells smaller than this size.

    Returns
    -------
    ndarray :
        3D uint8 array (2D image of 3 channels) containing the overlaid image.
    ndarray :
        3D uint8 array (2D image of 3 channels) containing the refined segmentation image.
    dict :
        Dictionary with scoring and settings information.
    """

    orig = np.array(orig)
    mask = np.full(orig.shape[0:2], LABEL_UNKNOWN, dtype=np.uint8)
    num_pos, num_neg = 0, 0

    if size_thresh is None:
        size_thresh = 0
    elif size_thresh == 'default':
        size_thresh = data['settings']['default_size_thresh']
    if marker_thresh == 'default':
        marker_thresh = data['settings']['default_marker_thresh']

    for cell in data['cells']:
        if data['dataVersion'] == 4:
            c = decode_cell_data_v4(cell)
        else:
            c = cell
        if c['size'] > size_thresh and (size_thresh_upper is None or c['size'] < size_thresh_upper):
            if c['positive'] or (marker_thresh is not None and c['marker'] > marker_thresh):
                num_pos += 1
                label = LABEL_BORDER_POS
            else:
                num_neg += 1
                label = LABEL_BORDER_NEG
            border = make_full_contour(c['boundary'])
            for b in border:
                mask[b[1], b[0]] = label

    mark_background(mask)
    fill_cells(mask)

    create_outer_boundary(mask)
    enlarge_cell_boundaries(mask)
    enlarge_cell_boundaries(mask)
    overlay, refined = create_final_images(np.array(orig), mask)

    num_total = num_pos + num_neg
    scoring = {
        'num_total': num_total,
        'num_pos': num_pos,
        'num_neg': num_neg,
        'percent_pos': round(num_pos / num_total * 100, 1) if num_pos > 0 else 0,
        'seg_thresh': data['settings']['seg_thresh'],
        'size_thresh': size_thresh,
        'size_thresh_upper': size_thresh_upper,
        'marker_thresh': marker_thresh,
    }

    return overlay, refined, scoring



import torch
from packaging import version
from torch import nn
import itertools
from ..util.image_pool import ImagePool
from .base_model import BaseModel
from . import networks
from .networks import get_optimizer


class CycleGANModel(BaseModel):
    """
    This class implements the CycleGAN model, for learning image-to-image translation without paired data.

    The model training requires '--dataset_mode unaligned' dataset.
    By default, it uses a '--netG resnet_9blocks' ResNet generator,
    a '--netD basic' discriminator (PatchGAN introduced by pix2pix),
    and a least-square GANs objective ('--gan_mode lsgan').

    CycleGAN paper: https://arxiv.org/pdf/1703.10593.pdf
    """

    def __init__(self, opt):
        """Initialize the CycleGAN class.

        Parameters:
            opt (Option class)-- stores all the experiment flags; needs to be a subclass of BaseOptions
        """
        BaseModel.__init__(self, opt)
        self.mod_gen_no = self.opt.modalities_no
        if not hasattr(self.opt,'upsample'):
            self.opt.upsample = 'convtranspose'
        if not hasattr(self.opt,'label_smoothing'):
            self.opt.label_smoothing = 0
        
        use_spectral_norm = self.opt.norm == 'spectral'
        
        self.loss_G_weights = opt.loss_G_weights
        self.loss_D_weights = opt.loss_D_weights
        self.loss_cyc_weights = [1 / self.mod_gen_no] * self.mod_gen_no
        
        self.opt.lambda_identity = 0 # do not use lambda identity for the first trial
        # specify the training losses you want to print out. The training/test scripts will call <BaseModel.get_current_losses>
        self.loss_names = ['D_A', 'G_A', 'cycle_A', 'idt_A', 'D_B', 'G_B', 'cycle_B', 'idt_B']
        # specify the images you want to save/display. The training/test scripts will call <BaseModel.get_current_visuals>
        l_suffix = range(1, self.opt.modalities_no + 1)
        visual_names_A = [f'real_As_{i}' for i in l_suffix] + [f'fake_Bs_{i}' for i in l_suffix] + [f'rec_As_{i}' for i in l_suffix]
        visual_names_B = [f'real_Bs_{i}' for i in l_suffix] + [f'fake_As_{i}' for i in l_suffix] + [f'rec_Bs_{i}' for i in l_suffix]

        # if self.is_train and self.opt.lambda_identity > 0.0:  # if identity loss is used, we also visualize idt_B=G_A(B) ad idt_A=G_A(B)
        #     visual_names_A.append('idt_B')
        #     visual_names_B.append('idt_A')

        self.visual_names = visual_names_A + visual_names_B  # combine visualizations for A and B
        # specify the models you want to save to the disk. The training/test scripts will call <BaseModel.save_networks> and <BaseModel.load_networks>.
        if self.is_train:
            self.model_names = [f'GA_{i}' for i in l_suffix] + [f'GB_{i}' for i in l_suffix] + [f'DA_{i}' for i in l_suffix] + [f'DB_{i}' for i in l_suffix]
        else:  # during test time, only load Gs
            if self.opt.BtoA:
                self.model_names = [f'GB_{i}' for i in l_suffix] 
            else:
                self.model_names = [f'GA_{i}' for i in l_suffix] 

        # define networks (both Generators and discriminators)
        # The naming is different from those used in the paper.
        # Code (vs. paper): G_A (G), G_B (F), D_A (D_Y), D_B (D_X)
        if isinstance(opt.net_g, str):
            self.opt.net_g = [self.opt.net_g] * self.mod_gen_no
        
        if version.parse(torch.__version__) < version.parse('1.11.0'):
            self.netGA = list()
            self.netGB = list()
        else:
            self.netGA = nn.ModuleList()
            self.netGB = nn.ModuleList()
        
        for i in range(self.mod_gen_no):
            if self.is_train or not self.opt.BtoA:
                self.netGA.append(networks.define_G(self.opt.input_nc, self.opt.output_nc, self.opt.ngf, self.opt.net_g[i], self.opt.norm,
                                                 not self.opt.no_dropout, self.opt.init_type, self.opt.init_gain, self.gpu_ids, self.opt.padding, 
                                                upsample=self.opt.upsample))
            if self.is_train or self.opt.BtoA:
                self.netGB.append(networks.define_G(self.opt.output_nc, self.opt.input_nc, self.opt.ngf, self.opt.net_g[i], self.opt.norm,
                                                 not self.opt.no_dropout, self.opt.init_type, self.opt.init_gain, self.gpu_ids, self.opt.padding, 
                                                 upsample=self.opt.upsample))
        
        if self.is_train:  # define a discriminator; conditional GANs need to take both input and output images; Therefore, #channels for D is input_nc + output_nc
            if version.parse(torch.__version__) < version.parse('1.11.0'):
                self.netDA = list()
                self.netDB = list()
            else:
                self.netDA = nn.ModuleList()
                self.netDB = nn.ModuleList()
            
            for i in range(self.mod_gen_no):
                self.netDA.append(networks.define_D(self.opt.output_nc, self.opt.ndf, self.opt.net_d,
                                                 self.opt.n_layers_D, self.opt.norm, self.opt.init_type, self.opt.init_gain,
                                                 self.gpu_ids))
                self.netDB.append(networks.define_D(self.opt.input_nc, self.opt.ndf, self.opt.net_d,
                                                 self.opt.n_layers_D, self.opt.norm, self.opt.init_type, self.opt.init_gain,
                                                 self.gpu_ids))



        if self.is_train:
            if opt.lambda_identity > 0.0:  # only works when input and output images have the same number of channels
                assert(opt.input_nc == opt.output_nc)
            self.fake_A_pools = [ImagePool(opt.pool_size) for _ in range(self.opt.modalities_no)]  # create image buffer to store previously generated images
            self.fake_B_pools = [ImagePool(opt.pool_size) for _ in range(self.opt.modalities_no)]  # create image buffer to store previously generated images
            
            # define loss functions
            # label smoothing currently only applies to discriminator losses & generatoe of lsgan/vanilla
            self.criterionGAN = networks.GANLoss(opt.gan_mode, label_smoothing=self.opt.label_smoothing).to(self.device)  # define GAN loss.
            self.criterionCycle = torch.nn.L1Loss()
            self.criterionIdt = torch.nn.L1Loss()
            
            self.criterionSmoothL1 = torch.nn.SmoothL1Loss()
            self.criterionVGG = networks.VGGLoss().to(self.device)
            
            
            # initialize optimizers
            params = []
            for i in range(len(self.netGA)):
                params += list(self.netGA[i].parameters())
            for i in range(len(self.netGB)):
                params += list(self.netGB[i].parameters())
            try:
                self.optimizer_G = get_optimizer(opt.optimizer)(params, lr=opt.lr_g, betas=(opt.beta1, 0.999))
            except:
                print(f'betas are not used for optimizer torch.optim.{opt.optimizer} in generators')
                self.optimizer_G = get_optimizer(opt.optimizer)(params, lr=opt.lr_g)
            
            params = []        
            for i in range(len(self.netDA)):
                params += list(self.netDA[i].parameters())
            for i in range(len(self.netDB)):
                params += list(self.netDB[i].parameters())
            
            # a smaller learning rate for discriminators to postpone training failure due to discriminators quickly become too strong
            try:
                self.optimizer_D = get_optimizer(opt.optimizer)(params, lr=opt.lr_d, betas=(opt.beta1, 0.999))
            except:
                print(f'betas are not used for optimizer torch.optim.{opt.optimizer} in generators')
                self.optimizer_D = get_optimizer(opt.optimizer)(params, lr=opt.lr_d)
            
            
            self.optimizers.append(self.optimizer_G)
            self.optimizers.append(self.optimizer_D)

    def set_input(self, input):
        """Unpack input data from the dataloader and perform necessary pre-processing steps.

        Parameters:
            input (dict): include the data itself and its metadata information.

        The option 'direction' can be used to swap domain A and domain B.
        """
        self.real_As = [input['A'].to(self.device) for _ in range(self.opt.modalities_no)]
        self.real_Bs = [x.to(self.device) for x in input['Bs']]
        self.image_paths = input['A_paths']

    def forward(self):
        """
        Run forward pass; called by both functions <optimize_parameters> and <test>.
        During inference, some output list could be empty. For example, if only netGAs are loaded,
        there will not be valid elements in self.rec_As and self.fake_As.
        """
        self.fake_Bs = [netGA(real_A) for netGA, real_A in zip(self.netGA, self.real_As)]  # G_A(A)
        self.rec_As = [netGB(fake_B) for netGB, fake_B in zip(self.netGB, self.fake_Bs)]   # G_B(G_A(A))
        
        self.fake_As = [netGB(real_B) for netGB, real_B in zip(self.netGB, self.real_Bs)]  # G_B(B)
        self.rec_Bs = [netGA(fake_A) for netGA, fake_A in zip(self.netGA, self.fake_As)]   # G_A(G_B(B))
    
      
    def backward_D_basic(self, netD, real, fake, scale_factor=1):
        """Calculate GAN loss for the discriminator

        Parameters:
            netD (network)      -- the discriminator D
            real (tensor array) -- real images
            fake (tensor array) -- images generated by a generator

        Return the discriminator loss.
        We also call loss_D.backward() to calculate the gradients.
        """
        # Real
        pred_real = netD(real)
        loss_D_real = self.criterionGAN(pred_real, True)
        # Fake
        pred_fake = netD(fake.detach())
        loss_D_fake = self.criterionGAN(pred_fake, False)
        # Combined loss and calculate gradients
        loss_D = (loss_D_real + loss_D_fake) * 0.5 * scale_factor
        loss_D.backward()
        return loss_D

    def backward_D_A(self):
        """Calculate GAN loss for discriminator D_A"""
        fake_Bs = [fake_B_pool.query(fake_B) for fake_B_pool, fake_B in zip(self.fake_B_pools, self.fake_Bs)]
        real_Bs = self.real_Bs
        
        self.loss_D_A = 0
        for i, (netDA, real_B, fake_B) in enumerate(zip(self.netDA, real_Bs, fake_Bs)):
            self.loss_D_A += self.backward_D_basic(netDA, real_B, fake_B, scale_factor=self.loss_D_weights[i])
        #self.loss_D_A.backward()

    def backward_D_B(self):
        """Calculate GAN loss for discriminator D_B"""
        fake_As = [fake_A_pool.query(fake_A) for fake_A_pool, fake_A in zip(self.fake_A_pools, self.fake_As)]
        real_As = self.real_As
            
        self.loss_D_B = 0
        for i, (netDB, real_A, fake_A) in enumerate(zip(self.netDB, real_As, fake_As)):
            self.loss_D_B += self.backward_D_basic(netDB, real_A, fake_A, scale_factor=self.loss_D_weights[i]) 
        #self.loss_D_B.backward()

    def backward_G(self):
        """Calculate the loss for generators G_A and G_B"""
        # default lambda values from cyclegan implementation:
        # https://github.com/junyanz/pytorch-CycleGAN-and-pix2pix/blob/c3268edd50ec37a81600c9b981841f48929671b8/models/cycle_gan_model.py#L41
        lambda_idt = 0#self.opt.lambda_identity # identity loss is used to preserve color consistency between input and output images, which we do not want to encourage
        lambda_A = 10#self.opt.lambda_A 
        lambda_B = 10#self.opt.lambda_B
        # Identity loss
        if lambda_idt > 0:
            # G_A should be identity if real_B is fed: ||G_A(B) - B||
            self.idt_A = self.netG_A(self.real_B)
            self.loss_idt_A = self.criterionIdt(self.idt_A, self.real_B) * lambda_B * lambda_idt
            # G_B should be identity if real_A is fed: ||G_B(A) - A||
            self.idt_B = self.netG_B(self.real_A)
            self.loss_idt_B = self.criterionIdt(self.idt_B, self.real_A) * lambda_A * lambda_idt
        else:
            self.loss_idt_A = 0
            self.loss_idt_B = 0
            
        # GAN loss D_A(G_A(A))
        self.loss_G_A = 0
        for i, (netDA, fake_B, real_B) in enumerate(zip(self.netDA, self.fake_Bs, self.real_Bs)):
            self.loss_G_A += self.criterionGAN(netDA(fake_B), True) * self.loss_G_weights[i]
            self.loss_G_A += self.criterionVGG(fake_B, real_B) * self.loss_G_weights[i]
        
        # GAN loss D_B(G_B(B))
        self.loss_G_B = 0
        for i, (netDB, fake_A, real_A) in enumerate(zip(self.netDB, self.fake_As, self.real_As)):
            self.loss_G_B += self.criterionGAN(netDB(fake_A), True) * self.loss_G_weights[i]
            self.loss_G_B += self.criterionVGG(fake_A, real_A) * self.loss_G_weights[i]
            
        # Forward cycle loss || G_B(G_A(A)) - A||
        self.loss_cycle_A = 0
        for i, (rec_A, real_A) in enumerate(zip(self.rec_As, self.real_As)):
            self.loss_cycle_A += self.criterionCycle(rec_A, real_A) * lambda_A * self.loss_cyc_weights[i]
        # Backward cycle loss || G_A(G_B(B)) - B||
        self.loss_cycle_B = 0
        for i, (rec_B, real_B) in enumerate(zip(self.rec_Bs, self.real_Bs)):
            self.loss_cycle_B += self.criterionCycle(rec_B, real_B) * lambda_B * self.loss_cyc_weights[i]
        
        # VGG loss
        # self.loss_G_VGG = self.criterionVGG(self.fake_B_1, self.real_B_1) * self.opt.lambda_feat
        
        # smooth L1
        # self.loss_G_A_L1 = self.criterionSmoothL1(self.fake_B_1, self.real_B_1) * self.opt.lambda_L1
        
        # combined loss and calculate gradients
        self.loss_G = self.loss_G_A + self.loss_G_B + self.loss_cycle_A + self.loss_cycle_B + self.loss_idt_A + self.loss_idt_B
        self.loss_G.backward()

    def optimize_parameters(self):
        """Calculate losses, gradients, and update network weights; called in every training iteration"""
        # forward
        self.forward()      # compute fake images and reconstruction images.
        # G_A and G_B
        self.set_requires_grad(self.netDA + self.netDB, False)  # Ds require no gradients when optimizing Gs
        self.optimizer_G.zero_grad()  # set G_A and G_B's gradients to zero
        self.backward_G()             # calculate gradients for G_A and G_B
        self.optimizer_G.step()       # update G_A and G_B's weights
        
        # D_A and D_B
        self.set_requires_grad(self.netDA + self.netDB, True)
        self.optimizer_D.zero_grad()   # set D_A and D_B's gradients to zero
        self.backward_D_A()      # calculate gradients for D_A
        self.backward_D_B()      # calculate graidents for D_B
        self.optimizer_D.step()  # update D_A and D_B's weights



"""This package contains modules related to objective functions, optimizations, and network architectures.

To add a custom model class called 'dummy', you need to add a file called 'dummy_model.py' and define a subclass DummyModel inherited from BaseModel.
You need to implement the following five functions:
    -- <__init__>:                      initialize the class; first call BaseModel.__init__(self, opt).
    -- <set_input>:                     unpack data from dataset and apply preprocessing.
    -- <forward>:                       produce intermediate results.
    -- <optimize_parameters>:           calculate loss, gradients, and update network weights.
    -- <modify_commandline_options>:    (optionally) add model-specific options and set default options.

In the function <__init__>, you need to define four lists:
    -- self.loss_names (str list):          specify the training losses that you want to plot and save.
    -- self.model_names (str list):         define networks used in our training.
    -- self.visual_names (str list):        specify the images that you want to display and save.
    -- self.optimizers (optimizer list):    define and initialize optimizers. You can define one optimizer for each network. If two networks are updated at the same time, you can use itertools.chain to group them. See CycleGAN_model.py for an usage.

Now you can use the model class by specifying flag '--model dummy'.
See our template model class 'template_model.py' for more details.
"""
import base64
import os
import itertools
import importlib
from functools import lru_cache
from io import BytesIO
import json
import math
import importlib.metadata
import pathlib

import requests
import torch
from PIL import Image
Image.MAX_IMAGE_PIXELS = None

import numpy as np
from dask import delayed, compute

from deepliif.util import *
from deepliif.util.util import tensor_to_pil
from deepliif.data import transform
from deepliif.postprocessing import compute_final_results, compute_cell_results, to_array
from deepliif.postprocessing import encode_cell_data_v4, decode_cell_data_v4
from deepliif.options import Options, print_options

from .base_model import BaseModel

# import for init purpose, not used in this script
from .DeepLIIF_model import DeepLIIFModel
from .DeepLIIFExt_model import DeepLIIFExtModel


@lru_cache
def get_opt(model_dir, mode='test'):
    """
    mode: test or train, currently only functions used for inference utilize get_opt so it
          defaults to test
    """
    if mode == 'train':
        opt = Options(path_file=os.path.join(model_dir,'train_opt.txt'), mode=mode)
    elif mode == 'test':
        try:
            opt = Options(path_file=os.path.join(model_dir,'test_opt.txt'), mode=mode)
        except:
            opt = Options(path_file=os.path.join(model_dir,'train_opt.txt'), mode=mode)
        opt.use_dp = False
        opt.gpu_ids = list(range(torch.cuda.device_count()))
    return opt


def find_model_using_name(model_name):
    """Import the module "models/[model_name]_model.py".

    In the file, the class called DatasetNameModel() will
    be instantiated. It has to be a subclass of BaseModel,
    and it is case-insensitive.
    """
    model_filename = "deepliif.models." + model_name + "_model"
    modellib = importlib.import_module(model_filename)
    model = None
    target_model_name = model_name.replace('_', '') + 'model'
    for name, cls in modellib.__dict__.items():
        if name.lower() == target_model_name.lower() \
                and issubclass(cls, BaseModel):
            model = cls

    if model is None:
        print("In %s.py, there should be a subclass of BaseModel with class name that matches %s in lowercase." % (
            model_filename, target_model_name))
        exit(0)

    return model


def get_option_setter(model_name):
    """Return the static method <modify_commandline_options> of the model class."""
    model_class = find_model_using_name(model_name)
    return model_class.modify_commandline_options


def create_model(opt):
    """Create a model given the option.

    This function warps the class CustomDatasetDataLoader.
    This is the main interface between this package and 'train.py'/'test.py'

    Example:
        >>> from deepliif.models import create_model
        >>> model = create_model(opt)
    """
    model = find_model_using_name(opt.model)
    instance = model(opt)
    print("model [%s] was created" % type(instance).__name__)
    return instance


def load_torchscript_model(model_pt_path, device):
    net = torch.jit.load(model_pt_path, map_location=device)
    net = disable_batchnorm_tracking_stats(net)
    net.eval()
    return net



def load_eager_models(opt, devices=None):
    # create a model given model and other options
    model = create_model(opt)
    # regular setup: load and print networks; create schedulers
    model.setup(opt)

    nets = {}
    if devices:
        model_names = list(devices.keys())
    else:
        model_names = model.model_names

    for name in model_names:#model.model_names:
        if isinstance(name, str):
            if '_' in name:
                net = getattr(model, 'net' + name.split('_')[0])[int(name.split('_')[-1]) - 1]
            else:
                net = getattr(model, 'net' + name)
    
            if opt.phase != 'train':
                net.eval()
                net = disable_batchnorm_tracking_stats(net)
            
            # SDG models when loaded are still DP.. not sure why
            if isinstance(net, torch.nn.DataParallel):
                net = net.module

            nets[name] = net
            if devices:
                nets[name].to(devices[name])
            
    return nets

@lru_cache
def init_nets(model_dir, eager_mode=False, opt=None, phase='test'):
    """
    Init DeepLIIF networks so that every net in
    the same group is deployed on the same GPU
    
    opt_args: to overwrite opt arguments in train_opt.txt, typically used in inference stage
              for example, opt_args={'phase':'test'}
    """ 
    if opt is None:
        opt = get_opt(model_dir, mode=phase)
    opt.use_dp = False
    
    # get l_net_groups: a list of lists, each sublist is a candidate, ordered by priority (try 1st, if fail try 2nd, etc.)
    if opt.model in ['DeepLIIF','DeepLIIFKD']:
        # net_groups = [
        #     ('G1', 'G52'),
        #     ('G2', 'G53'),
        #     ('G3', 'G54'),
        #     ('G4', 'G55'),
        #     ('G51',)
        # ]
        if opt.modalities_no == 0:
            net_groups = [(f'G{opt.mod_id_seg}{opt.input_id}',)]
        else:
            net_groups = [(f'G{i+1}', f'G{opt.mod_id_seg}{int(opt.input_id)+i+1}') for i in range(opt.modalities_no)]
            net_groups += [(f'G{opt.mod_id_seg}{opt.input_id}',)] # this is the generator for the input base mod
    elif opt.model in ['DeepLIIFExt','SDG']:
        if opt.seg_gen:
            net_groups = [(f'G_{i+1}',f'GS_{i+1}') for i in range(opt.modalities_no)]
        else:
            net_groups = [(f'G_{i+1}',) for i in range(opt.modalities_no)]
    elif opt.model == 'CycleGAN':
        if opt.BtoA:
            net_groups = [(f'GB_{i+1}',) for i in range(opt.modalities_no)]
        else:
            net_groups = [(f'GA_{i+1}',) for i in range(opt.modalities_no)]
    else:
        raise Exception(f'init_nets() not implemented for model {opt.model}')

    number_of_gpus_all = torch.cuda.device_count()
    number_of_gpus = min(len(opt.gpu_ids),number_of_gpus_all)

    if number_of_gpus > 0:
        mapping_gpu_ids = {i:idx for i,idx in enumerate(opt.gpu_ids)}
        chunks = [itertools.chain.from_iterable(c) for c in chunker(net_groups, number_of_gpus)]
        # chunks = chunks[1:]
        #l_devices = [{n: torch.device(f'cuda:{mapping_gpu_ids[i]}') for i, g in enumerate(chunks) for n in g} for chunks in l_chunks]
        devices = {n: torch.device(f'cuda:{i}') for i, g in enumerate(chunks) for n in g}
    else:
        devices = {n: torch.device('cpu') for n in itertools.chain.from_iterable(net_groups)}

    if eager_mode:
        return load_eager_models(opt, devices)

    return {
        n: load_torchscript_model(os.path.join(model_dir, f'{n}.pt'), device=d)
        for n, d in devices.items()
    }


def compute_overlap(img_size, tile_size):
    w, h = img_size
    if round(w / tile_size) == 1 and round(h / tile_size) == 1:
        return 0

    return tile_size // 4


def run_torchserve(img, model_path=None, nets=None, eager_mode=False, opt=None, seg_only=False, mod_only=False, seg_weights=None, use_dask=True, output_tensor=False):
    """
    eager_mode: not used in this function; put in place to be consistent with run_dask
           so that run_wrapper() could call either this function or run_dask with
           same syntax
    opt: same as eager_mode
    seg_only: same as eager_mode
    mod_only: same as eager_mode
    seg_weights: same as eager_mode
    nets: same as eager_mode
    """
    buffer = BytesIO()
    torch.save(transform(img.resize((opt.scale_size, opt.scale_size))), buffer)

    torchserve_host = os.getenv('TORCHSERVE_HOST', 'http://localhost')
    res = requests.post(
        f'{torchserve_host}/wfpredict/deepliif',
        json={'img': base64.b64encode(buffer.getvalue()).decode('utf-8')}
    )

    res.raise_for_status()

    def deserialize_tensor(bs):
        return torch.load(BytesIO(base64.b64decode(bs.encode())), map_location=torch.device('cpu'))

    return {k: tensor_to_pil(deserialize_tensor(v)) for k, v in res.json().items()}


def run_dask(img, model_path=None, nets=None, eager_mode=False, opt=None, seg_only=False, mod_only=False,
             seg_weights=None, use_dask=True, output_tensor=False):
    """
    Provide either the model path or the networks object.
    
    `eager_mode` is only applicable if model_path is provided.
    """
    assert model_path is not None or nets is not None, 'Provide either the model path or the networks object.'
    if nets is None:
        model_dir = os.getenv('DEEPLIIF_MODEL_DIR', model_path)
        nets = init_nets(model_dir, eager_mode, opt)
    
    if use_dask: # check if use_dask should be overwritten
        use_dask = True if opt.norm != 'spectral' else False
    
    if isinstance(img,torch.Tensor): # if img input is already a tensor, pass
        ts = img
    else:
        if opt.input_no > 1 or opt.model == 'SDG':
            l_ts = [transform(img_i.resize((opt.scale_size,opt.scale_size))) for img_i in img]
            ts = torch.cat(l_ts, dim=1)
        else:
            ts = transform(img.resize((opt.scale_size, opt.scale_size)))
    
    
    if use_dask:
        @delayed
        def forward(input, model):
            with torch.no_grad():
                return model(input.to(next(model.parameters()).device))
    else: # some train settings like spectral norm somehow in inference mode is not compatible with dask
        def forward(input, model):
            with torch.no_grad():
                return model(input.to(next(model.parameters()).device))
    
    if opt.model in ['DeepLIIF','DeepLIIFKD']:
        if seg_weights is None:
            # weights = {
            #     'G51': 0.5, # IHC
            #     'G52': 0.0, # Hema
            #     'G53': 0.0, # DAPI
            #     'G54': 0.0, # Lap2
            #     'G55': 0.5, # Marker
            # }
            weights = {f'G{opt.mod_id_seg}{int(opt.input_id)+i}': 1/(opt.modalities_no+1) for i in range(opt.modalities_no+1)}
        else:
            # weights = {
            #     'G51': seg_weights[0], # IHC
            #     'G52': seg_weights[1], # Hema
            #     'G53': seg_weights[2], # DAPI
            #     'G54': seg_weights[3], # Lap2
            #     'G55': seg_weights[4], # Marker
            # }
            weights = {f'G{opt.mod_id_seg}{int(opt.input_id)+i}': seg_weight for i,seg_weight in enumerate(seg_weights)}
        
        
        seg_map = {f'G{i+1}': f'G{opt.mod_id_seg}{int(opt.input_id)+i+1}' for i in range(opt.modalities_no)}
        if seg_only:
            seg_map = {k: v for k, v in seg_map.items() if weights[v] != 0}
        
        lazy_gens = {k: forward(ts, nets[k]) for k in seg_map}
        if 'Marker' in opt.modalities_names:
            mod_id_marker = opt.modalities_names.index("Marker")
            if f'G{mod_id_marker}' not in seg_map:
                lazy_gens[f'G{mod_id_marker}'] = forward(ts, nets[f'G{mod_id_marker}'])
        
        gens = compute(lazy_gens)[0]
        
        if not mod_only:
            lazy_segs = {v: forward(gens[k], nets[v]) for k, v in seg_map.items()}
            # run seg generator for the base input
            if weights[f'G{opt.mod_id_seg}{opt.input_id}'] != 0:
                lazy_segs[f'G{opt.mod_id_seg}{opt.input_id}'] = forward(ts, nets[f'G{opt.mod_id_seg}{opt.input_id}'])
            segs = compute(lazy_segs)[0]
        
            model_name_first = list(nets.keys())[0]
            device = next(nets[model_name_first].parameters()).device # take the device of the first net and move all outputs there for seg aggregation
            seg = torch.stack([torch.mul(segs[k].to(device), weights[k]) for k in segs.keys()]).sum(dim=0)
        
        if output_tensor:
            if mod_only:
                res = gens
            elif seg_only and opt.modalities_no > 0:
                res = {f'G{opt.modalities_no}': gens[f'G{opt.modalities_no}']} if f'G{opt.modalities_no}' in gens else {}
                res[f'G{opt.mod_id_seg}'] = seg
            else:
                res = {**gens, **segs}
                res[f'G{opt.mod_id_seg}'] = seg
            
        else:
            if mod_only:
                res = {k: tensor_to_pil(v.to(torch.device('cpu'))) for k, v in gens.items()}
            elif seg_only and opt.modalities_no > 0:
                res = {f'G{opt.modalities_no}': tensor_to_pil(gens[f'G{opt.modalities_no}'].to(torch.device('cpu')))} if f'G{opt.modalities_no}' in gens else {}
                res[f'G{opt.mod_id_seg}'] = tensor_to_pil(seg.to(torch.device('cpu')))
            else:
                res = {k: tensor_to_pil(v.to(torch.device('cpu'))) for k, v in gens.items()}
                res.update({k: tensor_to_pil(v.to(torch.device('cpu'))) for k, v in segs.items()})
                res[f'G{opt.mod_id_seg}'] = tensor_to_pil(seg.to(torch.device('cpu')))
    
        return res
    elif opt.model in ['DeepLIIFExt','SDG','CycleGAN']:
        if opt.model == 'CycleGAN':
            seg_map = {f'GB_{i+1}':None for i in range(opt.modalities_no)} if opt.BtoA else {f'GA_{i+1}':None for i in range(opt.modalities_no)}
        else:
            seg_map = {'G_' + str(i): 'GS_' + str(i) for i in range(1, opt.modalities_no + 1)}
        
        if use_dask:
            lazy_gens = {k: forward(ts, nets[k]) for k in seg_map}
            gens = compute(lazy_gens)[0]
        else:
            gens = {k: forward(ts, nets[k]) for k in seg_map}
        
        res = {k: tensor_to_pil(v) for k, v in gens.items()}
        if mod_only:
            return res
    
        if opt.seg_gen:
            if use_dask:
                lazy_segs = {v: forward(torch.cat([ts.to(torch.device('cpu')), gens[next(iter(seg_map))].to(torch.device('cpu')), gens[k].to(torch.device('cpu'))], 1), nets[v]).to(torch.device('cpu')) for k, v in seg_map.items()}
                segs = compute(lazy_segs)[0]
            else:
                segs = {v: forward(torch.cat([ts.to(torch.device('cpu')), gens[next(iter(seg_map))].to(torch.device('cpu')), gens[k].to(torch.device('cpu'))], 1), nets[v]).to(torch.device('cpu')) for k, v in seg_map.items()}
            res.update({k: tensor_to_pil(v) for k, v in segs.items()})
    
        return res
    else:
        raise Exception(f'run_dask() not fully implemented for {opt.model}')


def is_empty(tile):
    thresh = 9
    if isinstance(tile, list): # for pair of tiles, only mark it as empty / no need for prediction if ALL tiles are empty
        return all([True if image_variance_gray(t) < thresh else False for t in tile])
    else:
        return True if image_variance_gray(tile) < thresh else False


def run_wrapper(tile, run_fn, model_path=None, nets=None, eager_mode=False, opt=None, seg_only=False, mod_only=False, seg_weights=None, use_dask=True, output_tensor=False):
    if opt.model in ['DeepLIIF','DeepLIIFKD']:
        if is_empty(tile):
            if seg_only: # return seg image and the last translated modality
                res = {
                    #f'G{opt.modalities_no}': Image.new(mode='RGB', size=(512, 512), color=(10, 10, 10)),
                    f'G{opt.modalities_no}': Image.new(mode='RGB', size=(512, 512), color=opt.background_colors[-1]),
                    f'G{opt.mod_id_seg}': Image.new(mode='RGB', size=(512, 512), color=(0, 0, 0)),
                }
            elif mod_only:
                res = {f'G{i+1}': Image.new(mode='RGB', size=(512, 512), color=opt.background_colors[i]) for i in range(opt.modalities_no)}
                
            else :
                # return {
                #     'G1': Image.new(mode='RGB', size=(512, 512), color=(201, 211, 208)),
                #     'G2': Image.new(mode='RGB', size=(512, 512), color=(10, 10, 10)),
                #     'G3': Image.new(mode='RGB', size=(512, 512), color=(0, 0, 0)),
                #     'G4': Image.new(mode='RGB', size=(512, 512), color=(10, 10, 10)),
                #     'G5': Image.new(mode='RGB', size=(512, 512), color=(0, 0, 0)),
                #     'G51': Image.new(mode='RGB', size=(512, 512), color=(0, 0, 0)),
                #     'G52': Image.new(mode='RGB', size=(512, 512), color=(0, 0, 0)),
                #     'G53': Image.new(mode='RGB', size=(512, 512), color=(0, 0, 0)),
                #     'G54': Image.new(mode='RGB', size=(512, 512), color=(0, 0, 0)),
                #     'G55': Image.new(mode='RGB', size=(512, 512), color=(0, 0, 0)),
                # }
                res = {**{f'G{i+1}': Image.new(mode='RGB', size=(512, 512), color=opt.background_colors[i]) for i in range(opt.modalities_no)},
                       **{f'G{opt.mod_id_seg}': Image.new(mode='RGB', size=(512, 512), color=(0, 0, 0))}}
                
                # assign mod-wise seg output to corresponding keys, again currently input_id only supports 0 or 1
                if opt.input_id == 1:
                    res = {**res,
                           **{f'G{opt.mod_id_seg}{i+1}': Image.new(mode='RGB', size=(512, 512), color=(0, 0, 0)) for i in range(opt.modalities_no+1)}}
                else:
                    res = {**res,
                           **{f'G{opt.mod_id_seg}{i}': Image.new(mode='RGB', size=(512, 512), color=(0, 0, 0)) for i in range(opt.modalities_no+1)}}

            # when modalities_no = 0... we do not need to generate G0 - output should be just a seg image
            if 'G0' in res:
                del res['G0']
            return res
        else:
            return run_fn(tile, model_path, None, eager_mode, opt, seg_only, mod_only, seg_weights)
    elif opt.model in ['DeepLIIFExt', 'SDG']:
        if is_empty(tile):
            res = {'G_' + str(i): Image.new(mode='RGB', size=(512, 512)) for i in range(1, opt.modalities_no + 1)}
            res.update({'GS_' + str(i): Image.new(mode='RGB', size=(512, 512)) for i in range(1, opt.modalities_no + 1)})
            return res
        else:
            return run_fn(tile, model_path, None, eager_mode, opt)
    elif opt.model in ['CycleGAN']:
        if is_empty(tile):
            net_names = ['GB_{i+1}' for i in range(opt.modalities_no)] if opt.BtoA else [f'GA_{i+1}' for i in range(opt.modalities_no)]
            res = {net_name: Image.new(mode='RGB', size=(512, 512)) for net_name in net_names}
            return res
        else:
            return run_fn(tile, model_path, None, eager_mode, opt)
    else:
        raise Exception(f'run_wrapper() not implemented for model {opt.model}')


def inference(img, tile_size, overlap_size, model_path, use_torchserve=False,
              eager_mode=False, color_dapi=False, color_marker=False, opt=None,
              return_seg_intermediate=False, seg_only=False, mod_only=False,
              seg_weights=None, opt_args={}):
    """
    opt_args: a dictionary of key and values to add/overwrite to opt
    """
    if not opt:
        opt = get_opt(model_path)
        #print_options(opt)
    
    for k,v in opt_args.items():
        setattr(opt,k,v)
    #print_options(opt)

    run_fn = run_torchserve if use_torchserve else run_dask

    if opt.model == 'SDG':
        # SDG could have multiple input images/modalities, hence the input could be a rectangle.
        # We split the input to get each modality image then create tiles for each set of input images.
        w, h = int(img.width / opt.input_no), img.height
        orig = [img.crop((w * i, 0, w * (i+1), h)) for i in range(opt.input_no)]
    else:
        # Otherwise expect a single input image, which is used directly.
        orig = img

    tiler = InferenceTiler(orig, tile_size, overlap_size)
    for tile in tiler:
        tiler.stitch(run_wrapper(tile, run_fn, model_path, None, eager_mode, opt, seg_only, mod_only, seg_weights))
        
    results = tiler.results()
    
    if opt.model in ['DeepLIIF','DeepLIIFKD']:
        # check if both the elements and the order are exactly the same
        l_modname = [f'mod{i+1}' for i in range(opt.modalities_no)]
        if l_modname != opt.modalities_names[1:]: 
            # if not, append modalities_names to mod names
            l_modname = [f'mod{i+1}-{mod_name}' for i,mod_name in enumerate(opt.modalities_names[1:])]
        d_modname2id = {mod_name:f'G{i+1}' for i,mod_name in enumerate(l_modname)}
        
        if opt.seg_gen:
            l_modname_seg = [f'mod{i}' for i in range(opt.modalities_no+1)]
            if l_modname_seg != opt.modalities_names: 
                # if not, append modalities_names to mod names
                l_modname_seg = [f'mod{i}-{mod_name}' for i,mod_name in enumerate(opt.modalities_names)]
            if f'G{opt.mod_id_seg}0' in results.keys():
                d_modname2id_seg = {mod_name:f'G{opt.mod_id_seg}{i}' for i,mod_name in enumerate(l_modname_seg)}
            else:
                d_modname2id_seg = {mod_name:f'G{opt.mod_id_seg}{i+1}' for i,mod_name in enumerate(l_modname_seg)}
        
        if not mod_only:
            d_modname2id['Seg'] = f'G{opt.mod_id_seg}'
        
        #print('d_modname2id:',d_modname2id)
            
        if seg_only:
            images = {'Seg': results[d_modname2id['Seg']]}
            marker_key = find_marker_key(d_modname2id)
            if marker_key is not None:
                images[marker_key] = results[d_modname2id[marker_key]]
            # if f'G{opt.modalities_no}' in results:
            #     images.update({'Marker': results['G{opt.modalities_no}']})
            # if 'Marker' in d_modname2id:
            #     images.update({'Marker': results[d_modname2id['Marker']]})
        else:
            # images = {
            #     'Hema': results['G1'],
            #     'DAPI': results['G2'],
            #     'Lap2': results['G3'],
            #     'Marker': results['G4'],
            #     'Seg': results['G5'],
            # }
            # images = {f'mod{i+1}': results[f'G{i+1}'] for i in range(opt.modalities_no)}
            # images['Seg'] = results[f'G{opt.modalities_no+1}']
            images = {mod_name: results[mod_id] for mod_name,mod_id in d_modname2id.items()}
        
        if return_seg_intermediate and not seg_only:
            # images.update({'IHC_s':results['G51'],
            #               'Hema_s':results['G52'],
            #               'DAPI_s':results['G53'],
            #               'Lap2_s':results['G54'],
            #               'Marker_s':results['G55'],})
            
            #images.update({f'mod{i+1}_s':results[f'G{opt.modalities_no+1}{i+1}'] for i in range(opt.modalities_no+1)})
            #images.update({f'{mod_name}_s':results[f'G{opt.modalities_no+1}']})
            images.update({f'{mod_name}_s':results[d_modname2id_seg[mod_name]] for mod_name in d_modname2id_seg.keys()})
        
        # if color_dapi and not seg_only:
        #     matrix = (       0,        0,        0, 0,
        #               299/1000, 587/1000, 114/1000, 0,
        #               299/1000, 587/1000, 114/1000, 0)
        #     images['DAPI'] = images['DAPI'].convert('RGB', matrix)
        # if color_marker and not seg_only:
        #     matrix = (299/1000, 587/1000, 114/1000, 0,
        #               299/1000, 587/1000, 114/1000, 0,
        #                      0,        0,        0, 0)
        #     images['Marker'] = images['Marker'].convert('RGB', matrix)
        return images

    elif opt.model == 'DeepLIIFExt':
        images = {f'mod{i}': results[f'G_{i}'] for i in range(1, opt.modalities_no + 1)}
        if opt.seg_gen:
            images.update({f'Seg{i}': results[f'GS_{i}'] for i in range(1, opt.modalities_no + 1)})
        return images

    elif opt.model == 'SDG':
        images = {f'mod{i}': results[f'G_{i}'] for i in range(1, opt.modalities_no + 1)}
        return images

    else:
        #raise Exception(f'inference() not implemented for model {opt.model}')
        return results # return result images with default key names (i.e., net names)


def postprocess(orig, images, tile_size, model, seg_thresh=120, size_thresh='default', marker_thresh=None, size_thresh_upper=None):
    if model in ['DeepLIIF','DeepLIIFKD']:
        resolution = '40x' if tile_size > 384 else ('20x' if tile_size > 192 else '10x')
        overlay, refined, scoring = compute_final_results(
            orig, images['Seg'], images.get(find_marker_key(images)), resolution,
            size_thresh, marker_thresh, size_thresh_upper, seg_thresh)
        processed_images = {}
        processed_images['SegOverlaid'] = Image.fromarray(overlay)
        processed_images['SegRefined'] = Image.fromarray(refined)
        return processed_images, scoring

    elif model in ['DeepLIIFExt','SDG']:
        resolution = '40x' if tile_size > 768 else ('20x' if tile_size > 384 else '10x')
        processed_images = {}
        scoring = {}
        for img_name in list(images.keys()):
            if 'Seg' in img_name:
                seg_img = images[img_name]
                overlay, refined, score = compute_final_results(
                    orig, images[img_name], None, resolution,
                    size_thresh, marker_thresh, size_thresh_upper, seg_thresh)
    
                processed_images[img_name + '_Overlaid'] = Image.fromarray(overlay)
                processed_images[img_name + '_Refined'] = Image.fromarray(refined)
                scoring[img_name] = score
        return processed_images, scoring

    else:
        raise Exception(f'postprocess() not implemented for model {model}')


def infer_modalities(img, tile_size, model_dir, eager_mode=False,
                     color_dapi=False, color_marker=False, opt=None,
                     return_seg_intermediate=False, seg_only=False, mod_only=False, seg_weights=None):
    """
    This function is used to infer modalities for the given image using a trained model.
    :param img: The input image.
    :param tile_size: The tile size.
    :param model_dir: The directory containing serialized model files.
    :return: The inferred modalities and the segmentation mask.
    """
    if opt is None:
        opt = get_opt(model_dir)
        opt.use_dp = False
        #print_options(opt)
    
    # for those with multiple input modalities, find the correct size to calculate overlap_size
    input_no = opt.input_no if hasattr(opt, 'input_no') else 1
    img_size = (img.size[0] / input_no, img.size[1]) # (width, height)

    images = inference(
        img,
        tile_size=tile_size,
        #overlap_size=compute_overlap(img_size, tile_size),
        overlap_size=tile_size//16,
        model_path=model_dir,
        eager_mode=eager_mode,
        color_dapi=color_dapi,
        color_marker=color_marker,
        opt=opt,
        return_seg_intermediate=return_seg_intermediate,
        seg_only=seg_only,
        mod_only=mod_only,
        seg_weights=seg_weights,
    )

    if not hasattr(opt,'seg_gen') or (hasattr(opt,'seg_gen') and opt.seg_gen): # the first condition accounts for old settings of deepliif; the second refers to deepliifext models
        if not mod_only:
            post_images, scoring = postprocess(img, images, tile_size, opt.model)
            images = {**images, **post_images}
            if seg_only:
                delete_keys = [k for k in images.keys() if 'Seg' not in k]
                for name in delete_keys:
                    del images[name]
            return images, scoring
        else:
            return images, None
    else:
        return images, None


def infer_results_for_wsi(input_dir, filename, output_dir, model_dir, tile_size, region_size=20000, color_dapi=False, color_marker=False, seg_intermediate=False, seg_only=False, seg_weights=None):
    """
    This function infers modalities and segmentation mask for the given WSI image. It

    :param input_dir: The directory containing the WSI.
    :param filename: The WSI name.
    :param output_dir: The directory for saving the inferred modalities.
    :param model_dir: The directory containing the serialized model files.
    :param tile_size: The tile size.
    :param region_size: The size of each individual region to be processed at once.
    :return:
    """
    basename, _ = os.path.splitext(filename)
    results_dir = os.path.join(output_dir, basename)
    if not os.path.exists(results_dir):
        os.makedirs(results_dir)
    size_x, size_y, size_z, size_c, size_t, pixel_type = get_information(os.path.join(input_dir, filename))
    rescale = (pixel_type != 'uint8')
    print(filename, size_x, size_y, size_z, size_c, size_t, pixel_type, flush=True)

    results = {}
    scoring = None

    # javabridge already set up from previous call to get_information()
    with bioformats.ImageReader(os.path.join(input_dir, filename)) as reader:
        start_x, start_y = 0, 0

        while start_x < size_x:
            while start_y < size_y:
                print(start_x, start_y, flush=True)
                region_XYWH = (start_x, start_y, min(region_size, size_x - start_x), min(region_size, size_y - start_y))
                region = reader.read(XYWH=region_XYWH, rescale=rescale)
                img = Image.fromarray((region * 255).astype(np.uint8)) if rescale else Image.fromarray(region)

                region_modalities, region_scoring = infer_modalities(img, tile_size, model_dir, color_dapi=color_dapi, color_marker=color_marker, return_seg_intermediate=seg_intermediate, seg_only=seg_only, seg_weights=seg_weights)
                if region_scoring is not None:
                    if scoring is None:
                        scoring = {
                            'num_pos': region_scoring['num_pos'],
                            'num_neg': region_scoring['num_neg'],
                        }
                    else:
                        scoring['num_pos'] += region_scoring['num_pos']
                        scoring['num_neg'] += region_scoring['num_neg']

                for name, img in region_modalities.items():
                    if name not in results:
                        results[name] = np.zeros((size_y, size_x, 3), dtype=np.uint8)
                    results[name][region_XYWH[1]: region_XYWH[1] + region_XYWH[3],
                    region_XYWH[0]: region_XYWH[0] + region_XYWH[2]] = np.array(img)
                start_y += region_size
            start_y = 0
            start_x += region_size

    # write_results_to_pickle_file(os.path.join(results_dir, "results.pickle"), results)
    # read_results_from_pickle_file(os.path.join(results_dir, "results.pickle"))

    for name, img in results.items():
        write_big_tiff_file(os.path.join(results_dir, f'{basename}_{name}.ome.tiff'), img, tile_size)

    if scoring is not None:
        scoring['num_total'] = scoring['num_pos'] + scoring['num_neg']
        scoring['percent_pos'] = round(scoring['num_pos'] / scoring['num_total'] * 100, 1) if scoring['num_pos'] > 0 else 0
        with open(os.path.join(results_dir, f'{basename}.json'), 'w') as f:
            json.dump(scoring, f, indent=2)


def get_wsi_resolution(filename):
    """
    Try to get the resolution (magnification) of the slide and
    the corresponding tile size to use by default for DeepLIIF.
    If it cannot be found, return (None, None) instead.

    Parameters
    ----------
    filename : str
        Full path to the file.

    Returns
    -------
    str :
        Magnification (objective power) from image metadata.
    int :
        Corresponding tile size for DeepLIIF.
    """

    init_javabridge_bioformats()
    metadata = bioformats.get_omexml_metadata(filename)

    mag = None
    try:
        omexml = bioformats.OMEXML(metadata)
        mag = omexml.instrument().Objective.NominalMagnification
    except Exception as e:
        fields = ['AppMag', 'NominalMagnification']
        try:
            for field in fields:
                idx = metadata.find(field)
                if idx >= 0:
                    for i in range(idx, len(metadata)):
                        if metadata[i].isdigit() or metadata[i] == '.':
                            break
                    for j in range(i, len(metadata)):
                        if not metadata[j].isdigit() and metadata[j] != '.':
                            break
                    if i == j:
                        continue
                    mag = metadata[i:j]
                    break
        except Exception as e:
            pass

    if mag is None:
        return None, None

    try:
        tile_size = round((float(mag) / 40) * 512)
        return mag, tile_size
    except Exception as e:
        return None, None


def infer_cells_for_wsi(filename, model_dir, tile_size, region_size=20000, version=3, print_log=False, seg_weights=None):
    """
    Perform inference on a slide and get the results individual cell data.

    Parameters
    ----------
    filename : str
        Full path to the file.
    model_dir : str
        Full path to the directory with the DeepLIIF model files.
    tile_size : int
        Size of tiles to extract and perform inference on.
    region_size : int
        Maximum size to split the slide for processing.
    version : int
        Version of cell data to return (3 or 4).
    print_log : bool
        Whether or not to print updates while processing.
    seg_weights : list | tuple
        Optional five weights to use for creating the combined Seg image.
        If None, then the default weights are used.

    Returns
    -------
    dict :
        Individual cell data and associated values.
    """

    def print_info(*args):
        if print_log:
            print(*args, flush=True)

    resolution = '40x' if tile_size > 384 else ('20x' if tile_size > 192 else '10x')

    data = None
    default_marker_thresh, count_marker_thresh = 0, 0
    default_size_thresh, count_size_thresh = 0, 0

    with WSIReader(filename) as reader:
        size_x = reader.width
        size_y = reader.height
        print_info('Info:', size_x, size_y)

        num_regions_x = math.ceil(size_x / region_size)
        num_regions_y = math.ceil(size_y / region_size)
        stride_x = math.ceil(size_x / num_regions_x)
        stride_y = math.ceil(size_y / num_regions_y)
        print_info('Strides:', stride_x, stride_y)

        start_x, start_y = 0, 0

        while start_y < size_y:
            while start_x < size_x:
                region_XYWH = (start_x, start_y, min(stride_x, size_x-start_x), min(stride_y, size_y-start_y))
                print_info('Region:', region_XYWH)

                region = reader.read(region_XYWH)
                print_info(region.shape, region.dtype)
                img = Image.fromarray(region)
                print_info(img.size, img.mode)
                del region

                images = inference(
                    img,
                    tile_size=tile_size,
                    overlap_size=tile_size//16,
                    model_path=model_dir,
                    eager_mode=False,
                    color_dapi=False,
                    color_marker=False,
                    opt=None,
                    return_seg_intermediate=False,
                    seg_only=True,
                    seg_weights=seg_weights,
                )
                del img

                seg = to_array(images['Seg'])
                del images['Seg']
                marker_key = find_marker_key(images)
                marker = to_array(images[marker_key], True) if marker_key is not None else None
                del images
                region_data = compute_cell_results(seg, marker, resolution, version=version)
                del seg
                del marker

                if start_x != 0 or start_y != 0:
                    for i in range(len(region_data['cells'])):
                        cell = decode_cell_data_v4(region_data['cells'][i]) if version == 4 else region_data['cells'][i]
                        for j in range(2):
                            cell['bbox'][j] = (cell['bbox'][j][0] + start_x, cell['bbox'][j][1] + start_y)
                        cell['centroid'] = (cell['centroid'][0] + start_x, cell['centroid'][1] + start_y)
                        for j in range(len(cell['boundary'])):
                            cell['boundary'][j] = (cell['boundary'][j][0] + start_x, cell['boundary'][j][1] + start_y)
                        region_data['cells'][i] = encode_cell_data_v4(cell) if version == 4 else cell

                if data is None:
                    data = region_data
                else:
                    data['cells'] += region_data['cells']

                if region_data['settings']['default_marker_thresh'] is not None and region_data['settings']['default_marker_thresh'] != 0:
                    default_marker_thresh += region_data['settings']['default_marker_thresh']
                    count_marker_thresh += 1
                if region_data['settings']['default_size_thresh'] != 0:
                    default_size_thresh += region_data['settings']['default_size_thresh']
                    count_size_thresh += 1

                start_x += stride_x

            start_x = 0
            start_y += stride_y

    if count_marker_thresh == 0:
        count_marker_thresh = 1
    if count_size_thresh == 0:
        count_size_thresh = 1
    data['settings']['default_marker_thresh'] = round(default_marker_thresh / count_marker_thresh)
    data['settings']['default_size_thresh'] = round(default_size_thresh / count_size_thresh)

    data['settings']['tile_size'] = tile_size
    data['settings']['region_size'] = region_size
    data['settings']['seg_weights'] = seg_weights

    try:
        data['deepliifVersion'] = importlib.metadata.version('deepliif')
    except Exception as e:
        data['deepliifVersion'] = 'unknown'

    try:
        data['modelVersion'] = pathlib.PurePath(model_dir).name
    except Exception as e:
        data['modelVersion'] = 'unknown'

    return data


def find_marker_key(dictionary):
    for key in dictionary:
        if key.endswith('Marker'):
            return key
    return None
